import importlib
import os
from typing import TYPE_CHECKING, Iterable, Optional, Union, overload

import moto

if TYPE_CHECKING:
    from typing_extensions import Literal

    from moto.acm.models import AWSCertificateManagerBackend
    from moto.acmpca.models import ACMPCABackend
    from moto.amp.models import PrometheusServiceBackend
    from moto.apigateway.models import APIGatewayBackend
    from moto.apigatewaymanagementapi.models import ApiGatewayManagementApiBackend
    from moto.apigatewayv2.models import ApiGatewayV2Backend
    from moto.appconfig.models import AppConfigBackend
    from moto.applicationautoscaling.models import ApplicationAutoscalingBackend
    from moto.appmesh.models import AppMeshBackend
    from moto.appsync.models import AppSyncBackend
    from moto.athena.models import AthenaBackend
    from moto.autoscaling.models import AutoScalingBackend
    from moto.awslambda.models import LambdaBackend
    from moto.batch.models import BatchBackend
    from moto.bedrock.models import BedrockBackend
    from moto.bedrockagent.models import AgentsforBedrockBackend
    from moto.budgets.models import BudgetsBackend
    from moto.ce.models import CostExplorerBackend
    from moto.cloudformation.models import CloudFormationBackend
    from moto.cloudfront.models import CloudFrontBackend
    from moto.cloudtrail.models import CloudTrailBackend
    from moto.cloudwatch.models import CloudWatchBackend
    from moto.codebuild.models import CodeBuildBackend
    from moto.codecommit.models import CodeCommitBackend
    from moto.codepipeline.models import CodePipelineBackend
    from moto.cognitoidentity.models import CognitoIdentityBackend
    from moto.cognitoidp.models import CognitoIdpBackend
    from moto.comprehend.models import ComprehendBackend
    from moto.config.models import ConfigBackend
    from moto.core.base_backend import SERVICE_BACKEND, BackendDict
    from moto.databrew.models import DataBrewBackend
    from moto.datapipeline.models import DataPipelineBackend
    from moto.datasync.models import DataSyncBackend
    from moto.dax.models import DAXBackend
    from moto.directconnect.models import DirectConnectBackend
    from moto.dms.models import DatabaseMigrationServiceBackend
    from moto.ds.models import DirectoryServiceBackend
    from moto.dsql.models import AuroraDSQLBackend
    from moto.dynamodb.models import DynamoDBBackend
    from moto.dynamodb_v20111205.models import (
        DynamoDBBackend as DynamoDBBackend_v20111205,
    )
    from moto.dynamodbstreams.models import DynamoDBStreamsBackend
    from moto.ebs.models import EBSBackend
    from moto.ec2.models import EC2Backend
    from moto.ec2instanceconnect.models import Ec2InstanceConnectBackend
    from moto.ecr.models import ECRBackend
    from moto.ecs.models import EC2ContainerServiceBackend
    from moto.efs.models import EFSBackend
    from moto.eks.models import EKSBackend
    from moto.elasticache.models import ElastiCacheBackend
    from moto.elasticbeanstalk.models import EBBackend
    from moto.elastictranscoder.models import ElasticTranscoderBackend
    from moto.elb.models import ELBBackend
    from moto.elbv2.models import ELBv2Backend
    from moto.emr.models import ElasticMapReduceBackend
    from moto.emrcontainers.models import EMRContainersBackend
    from moto.emrserverless.models import EMRServerlessBackend
    from moto.es.models import ElasticsearchServiceBackend
    from moto.events.models import EventsBackend
    from moto.firehose.models import FirehoseBackend
    from moto.forecast.models import ForecastBackend
    from moto.fsx.models import FSxBackend
    from moto.glacier.models import GlacierBackend
    from moto.glue.models import GlueBackend
    from moto.greengrass.models import GreengrassBackend
    from moto.guardduty.models import GuardDutyBackend
    from moto.iam.models import IAMBackend
    from moto.identitystore.models import IdentityStoreBackend
    from moto.inspector2.models import Inspector2Backend
    from moto.instance_metadata.models import InstanceMetadataBackend
    from moto.iot.models import IoTBackend
    from moto.iotdata.models import IoTDataPlaneBackend
    from moto.ivs.models import IVSBackend
    from moto.kafka.models import KafkaBackend
    from moto.kinesis.models import KinesisBackend
    from moto.kinesisvideo.models import KinesisVideoBackend
    from moto.kinesisvideoarchivedmedia.models import KinesisVideoArchivedMediaBackend
    from moto.kms.models import KmsBackend
    from moto.lakeformation.models import LakeFormationBackend
    from moto.lexv2models.models import LexModelsV2Backend
    from moto.logs.models import LogsBackend
    from moto.managedblockchain.models import ManagedBlockchainBackend
    from moto.mediaconnect.models import MediaConnectBackend
    from moto.medialive.models import MediaLiveBackend
    from moto.mediapackage.models import MediaPackageBackend
    from moto.mediastore.models import MediaStoreBackend
    from moto.mediastoredata.models import MediaStoreDataBackend
    from moto.memorydb.models import MemoryDBBackend
    from moto.meteringmarketplace.models import MeteringMarketplaceBackend
    from moto.moto_api._internal.models import MotoAPIBackend
    from moto.mq.models import MQBackend
    from moto.networkmanager.models import NetworkManagerBackend
    from moto.opensearch.models import OpenSearchServiceBackend
    from moto.opensearchserverless.models import OpenSearchServiceServerlessBackend
    from moto.opsworks.models import OpsWorksBackend
    from moto.organizations.models import OrganizationsBackend
    from moto.osis.models import OpenSearchIngestionBackend
    from moto.personalize.models import PersonalizeBackend
    from moto.pinpoint.models import PinpointBackend
    from moto.polly.models import PollyBackend
    from moto.qldb.models import QLDBBackend
    from moto.quicksight.models import QuickSightBackend
    from moto.ram.models import ResourceAccessManagerBackend
    from moto.rds.models import RDSBackend
    from moto.rdsdata.models import RDSDataServiceBackend
    from moto.redshift.models import RedshiftBackend
    from moto.redshiftdata.models import RedshiftDataAPIServiceBackend
    from moto.rekognition.models import RekognitionBackend
    from moto.resiliencehub.models import ResilienceHubBackend
    from moto.resourcegroups.models import ResourceGroupsBackend
    from moto.resourcegroupstaggingapi.models import ResourceGroupsTaggingAPIBackend
    from moto.robomaker.models import RoboMakerBackend
    from moto.route53.models import Route53Backend
    from moto.route53domains.models import Route53DomainsBackend
    from moto.route53resolver.models import Route53ResolverBackend
    from moto.s3.models import S3Backend
    from moto.s3control.models import S3ControlBackend
    from moto.s3tables.models import S3TablesBackend
    from moto.sagemaker.models import SageMakerModelBackend
    from moto.sagemakermetrics.models import SageMakerMetricsBackend
    from moto.sagemakerruntime.models import SageMakerRuntimeBackend
    from moto.scheduler.models import EventBridgeSchedulerBackend
    from moto.sdb.models import SimpleDBBackend
    from moto.secretsmanager.models import SecretsManagerBackend
    from moto.servicediscovery.models import ServiceDiscoveryBackend
    from moto.servicequotas.models import ServiceQuotasBackend
    from moto.ses.models import SESBackend
    from moto.sesv2.models import SESV2Backend
    from moto.shield.models import ShieldBackend
    from moto.signer.models import SignerBackend
    from moto.sns.models import SNSBackend
    from moto.sqs.models import SQSBackend
    from moto.ssm.models import SimpleSystemManagerBackend
    from moto.ssoadmin.models import SSOAdminBackend
    from moto.stepfunctions.models import StepFunctionBackend
    from moto.sts.models import STSBackend
    from moto.support.models import SupportBackend
    from moto.swf.models import SWFBackend
    from moto.textract.models import TextractBackend
    from moto.timestreaminfluxdb.models import TimestreamInfluxDBBackend
    from moto.timestreamquery.models import TimestreamQueryBackend
    from moto.timestreamwrite.models import TimestreamWriteBackend
    from moto.transcribe.models import TranscribeBackend
    from moto.transfer.models import TransferBackend
    from moto.wafv2.models import WAFV2Backend
    from moto.workspaces.models import WorkSpacesBackend
    from moto.workspacesweb.models import WorkSpacesWebBackend
    from moto.xray.models import XRayBackend


ALT_SERVICE_NAMES = {
    "lambda": "awslambda",
    "moto_api": "moto_api._internal",
    "neptune": "rds",
}
ALT_BACKEND_NAMES = {
    "moto_api._internal": "moto_api",
    "awslambda": "lambda",
    "awslambda_simple": "lambda_simple",
    "dynamodb_v20111205": "dynamodb",
    "elasticbeanstalk": "eb",
    "neptune": "rds",
}


def list_of_moto_modules() -> Iterable[str]:
    path = os.path.dirname(moto.__file__)
    for backend in sorted(os.listdir(path)):
        is_dir = os.path.isdir(os.path.join(path, backend))
        valid_folder = not backend.startswith("__")
        if is_dir and valid_folder:
            yield backend


def get_service_from_url(url: str) -> Optional[str]:
    from moto.backend_index import backend_url_patterns

    for service, pattern in backend_url_patterns:
        if pattern.match(url):
            return service
    return None


# There's a similar Union that we could import from boto3-stubs, but it wouldn't have
# moto's custom service backends
SERVICE_NAMES = Union[
    "Literal['acm']",
    "Literal['acm-pca']",
    "Literal['amp']",
    "Literal['apigateway']",
    "Literal['apigatewaymanagementapi']",
    "Literal['apigatewayv2']",
    "Literal['appconfig']",
    "Literal['applicationautoscaling']",
    "Literal['appmesh']",
    "Literal['appsync']",
    "Literal['athena']",
    "Literal['autoscaling']",
    "Literal['batch']",
    "Literal['bedrock']",
    "Literal['bedrock-agent']",
    "Literal['budgets']",
    "Literal['ce']",
    "Literal['cloudformation']",
    "Literal['cloudfront']",
    "Literal['cloudtrail']",
    "Literal['cloudwatch']",
    "Literal['codebuild']",
    "Literal['codecommit']",
    "Literal['codepipeline']",
    "Literal['cognito-identity']",
    "Literal['cognito-idp']",
    "Literal['comprehend']",
    "Literal['config']",
    "Literal['databrew']",
    "Literal['datapipeline']",
    "Literal['datasync']",
    "Literal['dax']",
    "Literal['directconnect']",
    "Literal['dms']",
    "Literal['ds']",
    "Literal['dsql']",
    "Literal['dynamodb']",
    "Literal['dynamodb_v20111205']",
    "Literal['dynamodbstreams']",
    "Literal['ebs']",
    "Literal['ec2']",
    "Literal['ec2instanceconnect']",
    "Literal['ecr']",
    "Literal['ecs']",
    "Literal['efs']",
    "Literal['eks']",
    "Literal['elasticache']",
    "Literal['elasticbeanstalk']",
    "Literal['elastictranscoder']",
    "Literal['elb']",
    "Literal['elbv2']",
    "Literal['emr']",
    "Literal['emr-containers']",
    "Literal['emr-serverless']",
    "Literal['es']",
    "Literal['events']",
    "Literal['firehose']",
    "Literal['forecast']",
    "Literal['fsx']",
    "Literal['glacier']",
    "Literal['glue']",
    "Literal['greengrass']",
    "Literal['guardduty']",
    "Literal['iam']",
    "Literal['identitystore']",
    "Literal['inspector2']",
    "Literal['instance_metadata']",
    "Literal['iot']",
    "Literal['iot-data']",
    "Literal['ivs']",
    "Literal['kafka']",
    "Literal['kinesis']",
    "Literal['kinesisvideo']",
    "Literal['kinesis-video-archived-media']",
    "Literal['kms']",
    "Literal['lakeformation']",
    "Literal['lambda']",
    "Literal['lexv2models']",
    "Literal['logs']",
    "Literal['managedblockchain']",
    "Literal['mediaconnect']",
    "Literal['medialive']",
    "Literal['mediapackage']",
    "Literal['mediastore']",
    "Literal['memorydb']",
    "Literal['mediastore-data']",
    "Literal['meteringmarketplace']",
    "Literal['moto_api']",
    "Literal['mq']",
    "Literal['neptune']",
    "Literal['networkmanager']",
    "Literal['opensearch']",
    "Literal['opensearchserverless']",
    "Literal['opsworks']",
    "Literal['organizations']",
    "Literal['osis']",
    "Literal['personalize']",
    "Literal['pinpoint']",
    "Literal['polly']",
    "Literal['qldb']",
    "Literal['quicksight']",
    "Literal['ram']",
    "Literal['rds']",
    "Literal['rds-data']",
    "Literal['redshift']",
    "Literal['redshift-data']",
    "Literal['rekognition']",
    "Literal['resiliencehub']",
    "Literal['resource-groups']",
    "Literal['resourcegroupstaggingapi']",
    "Literal['robomaker']",
    "Literal['route53']",
    "Literal['route53resolver']",
    "Literal['route53domains']",
    "Literal['s3']",
    "Literal['s3bucket_path']",
    "Literal['s3control']",
    "Literal['s3tables']",
    "Literal['sagemaker']",
    "Literal['sagemaker-metrics']",
    "Literal['sagemaker-runtime']",
    "Literal['scheduler']",
    "Literal['sdb']",
    "Literal['secretsmanager']",
    "Literal['servicediscovery']",
    "Literal['service-quotas']",
    "Literal['ses']",
    "Literal['sesv2']",
    "Literal['shield']",
    "Literal['signer']",
    "Literal['sns']",
    "Literal['sqs']",
    "Literal['ssm']",
    "Literal['sso-admin']",
    "Literal['stepfunctions']",
    "Literal['sts']",
    "Literal['support']",
    "Literal['swf']",
    "Literal['textract']",
    "Literal['timestream-influxdb']",
    "Literal['timestream-query']",
    "Literal['timestream-write']",
    "Literal['transcribe']",
    "Literal['transfer']",
    "Literal['wafv2']",
    "Literal['workspaces']",
    "Literal['workspaces-web']",
    "Literal['xray']",
]


def _import_backend(
    module_name: str,
    backends_name: str,
) -> "BackendDict[SERVICE_BACKEND]":
    module = importlib.import_module("moto." + module_name)
    return getattr(module, backends_name)


@overload
def get_backend(
    name: "Literal['acm']",
) -> "BackendDict[AWSCertificateManagerBackend]": ...
@overload
def get_backend(name: "Literal['acm-pca']") -> "BackendDict[ACMPCABackend]": ...
@overload
def get_backend(name: "Literal['amp']") -> "BackendDict[PrometheusServiceBackend]": ...
@overload
def get_backend(name: "Literal['apigateway']") -> "BackendDict[APIGatewayBackend]": ...
@overload
def get_backend(
    name: "Literal['apigatewaymanagementapi']",
) -> "BackendDict[ApiGatewayManagementApiBackend]": ...
@overload
def get_backend(
    name: "Literal['apigatewayv2']",
) -> "BackendDict[ApiGatewayV2Backend]": ...
@overload
def get_backend(name: "Literal['appconfig']") -> "BackendDict[AppConfigBackend]": ...
@overload
def get_backend(
    name: "Literal['applicationautoscaling']",
) -> "BackendDict[ApplicationAutoscalingBackend]": ...
@overload
def get_backend(name: "Literal['appmesh']") -> "BackendDict[AppMeshBackend]": ...
@overload
def get_backend(name: "Literal['appsync']") -> "BackendDict[AppSyncBackend]": ...
@overload
def get_backend(name: "Literal['athena']") -> "BackendDict[AthenaBackend]": ...
@overload
def get_backend(
    name: "Literal['autoscaling']",
) -> "BackendDict[AutoScalingBackend]": ...
@overload
def get_backend(name: "Literal['batch']") -> "BackendDict[BatchBackend]": ...
@overload
def get_backend(name: "Literal['bedrock']") -> "BackendDict[BedrockBackend]": ...
@overload
def get_backend(
    name: "Literal['bedrock-agent']",
) -> "BackendDict[AgentsforBedrockBackend]": ...
@overload
def get_backend(name: "Literal['budgets']") -> "BackendDict[BudgetsBackend]": ...
@overload
def get_backend(name: "Literal['ce']") -> "BackendDict[CostExplorerBackend]": ...
@overload
def get_backend(
    name: "Literal['cloudformation']",
) -> "BackendDict[CloudFormationBackend]": ...
@overload
def get_backend(name: "Literal['cloudfront']") -> "BackendDict[CloudFrontBackend]": ...
@overload
def get_backend(name: "Literal['cloudtrail']") -> "BackendDict[CloudTrailBackend]": ...
@overload
def get_backend(name: "Literal['cloudwatch']") -> "BackendDict[CloudWatchBackend]": ...
@overload
def get_backend(name: "Literal['codebuild']") -> "BackendDict[CodeBuildBackend]": ...
@overload
def get_backend(name: "Literal['codecommit']") -> "BackendDict[CodeCommitBackend]": ...
@overload
def get_backend(
    name: "Literal['codepipeline']",
) -> "BackendDict[CodePipelineBackend]": ...
@overload
def get_backend(
    name: "Literal['cognito-identity']",
) -> "BackendDict[CognitoIdentityBackend]": ...
@overload
def get_backend(name: "Literal['cognito-idp']") -> "BackendDict[CognitoIdpBackend]": ...
@overload
def get_backend(name: "Literal['comprehend']") -> "BackendDict[ComprehendBackend]": ...
@overload
def get_backend(name: "Literal['config']") -> "BackendDict[ConfigBackend]": ...
@overload
def get_backend(name: "Literal['databrew']") -> "BackendDict[DataBrewBackend]": ...
@overload
def get_backend(
    name: "Literal['datapipeline']",
) -> "BackendDict[DataPipelineBackend]": ...
@overload
def get_backend(name: "Literal['datasync']") -> "BackendDict[DataSyncBackend]": ...
@overload
def get_backend(name: "Literal['dax']") -> "BackendDict[DAXBackend]": ...
@overload
def get_backend(
    name: "Literal['dms']",
) -> "BackendDict[DatabaseMigrationServiceBackend]": ...
@overload
def get_backend(
    name: "Literal['directconnect']",
) -> "BackendDict[DirectConnectBackend]": ...
@overload
def get_backend(name: "Literal['ds']") -> "BackendDict[DirectoryServiceBackend]": ...
@overload
def get_backend(name: "Literal['dsql']") -> "BackendDict[AuroraDSQLBackend]": ...
@overload
def get_backend(name: "Literal['dynamodb']") -> "BackendDict[DynamoDBBackend]": ...
@overload
def get_backend(
    name: "Literal['dynamodb_v20111205']",
) -> "BackendDict[DynamoDBBackend_v20111205]": ...
@overload
def get_backend(
    name: "Literal['dynamodbstreams']",
) -> "BackendDict[DynamoDBStreamsBackend]": ...
@overload
def get_backend(name: "Literal['ebs']") -> "BackendDict[EBSBackend]": ...
@overload
def get_backend(name: "Literal['ec2']") -> "BackendDict[EC2Backend]": ...
@overload
def get_backend(
    name: "Literal['ec2instanceconnect']",
) -> "BackendDict[Ec2InstanceConnectBackend]": ...
@overload
def get_backend(name: "Literal['ecr']") -> "BackendDict[ECRBackend]": ...
@overload
def get_backend(
    name: "Literal['ecs']",
) -> "BackendDict[EC2ContainerServiceBackend]": ...
@overload
def get_backend(name: "Literal['efs']") -> "BackendDict[EFSBackend]": ...
@overload
def get_backend(name: "Literal['eks']") -> "BackendDict[EKSBackend]": ...
@overload
def get_backend(
    name: "Literal['elasticache']",
) -> "BackendDict[ElastiCacheBackend]": ...
@overload
def get_backend(name: "Literal['elasticbeanstalk']") -> "BackendDict[EBBackend]": ...
@overload
def get_backend(
    name: "Literal['elastictranscoder']",
) -> "BackendDict[ElasticTranscoderBackend]": ...
@overload
def get_backend(name: "Literal['elb']") -> "BackendDict[ELBBackend]": ...
@overload
def get_backend(name: "Literal['elbv2']") -> "BackendDict[ELBv2Backend]": ...
@overload
def get_backend(name: "Literal['emr']") -> "BackendDict[ElasticMapReduceBackend]": ...
@overload
def get_backend(
    name: "Literal['emr-containers']",
) -> "BackendDict[EMRContainersBackend]": ...
@overload
def get_backend(
    name: "Literal['emr-serverless']",
) -> "BackendDict[EMRServerlessBackend]": ...
@overload
def get_backend(
    name: "Literal['es']",
) -> "BackendDict[ElasticsearchServiceBackend]": ...
@overload
def get_backend(name: "Literal['events']") -> "BackendDict[EventsBackend]": ...
@overload
def get_backend(name: "Literal['firehose']") -> "BackendDict[FirehoseBackend]": ...
@overload
def get_backend(name: "Literal['forecast']") -> "BackendDict[ForecastBackend]": ...
@overload
def get_backend(name: "Literal['fsx']") -> "BackendDict[FSxBackend]": ...
@overload
def get_backend(name: "Literal['glacier']") -> "BackendDict[GlacierBackend]": ...
@overload
def get_backend(name: "Literal['glue']") -> "BackendDict[GlueBackend]": ...
@overload
def get_backend(name: "Literal['greengrass']") -> "BackendDict[GreengrassBackend]": ...
@overload
def get_backend(name: "Literal['guardduty']") -> "BackendDict[GuardDutyBackend]": ...
@overload
def get_backend(name: "Literal['iam']") -> "BackendDict[IAMBackend]": ...
@overload
def get_backend(
    name: "Literal['identitystore']",
) -> "BackendDict[IdentityStoreBackend]": ...
@overload
def get_backend(name: "Literal['inspector2']") -> "BackendDict[Inspector2Backend]": ...
@overload
def get_backend(
    name: "Literal['instance_metadata']",
) -> "BackendDict[InstanceMetadataBackend]": ...
@overload
def get_backend(name: "Literal['iot']") -> "BackendDict[IoTBackend]": ...
@overload
def get_backend(name: "Literal['iot-data']") -> "BackendDict[IoTDataPlaneBackend]": ...
@overload
def get_backend(name: "Literal['ivs']") -> "BackendDict[IVSBackend]": ...
@overload
def get_backend(name: "Literal['kafka']") -> "BackendDict[KafkaBackend]": ...
@overload
def get_backend(name: "Literal['kinesis']") -> "BackendDict[KinesisBackend]": ...
@overload
def get_backend(
    name: "Literal['kinesisvideo']",
) -> "BackendDict[KinesisVideoBackend]": ...
@overload
def get_backend(
    name: "Literal['kinesis-video-archived-media']",
) -> "BackendDict[KinesisVideoArchivedMediaBackend]": ...
@overload
def get_backend(name: "Literal['kms']") -> "BackendDict[KmsBackend]": ...
@overload
def get_backend(
    name: "Literal['lakeformation']",
) -> "BackendDict[LakeFormationBackend]": ...
@overload
def get_backend(name: "Literal['lambda']") -> "BackendDict[LambdaBackend]": ...
@overload
def get_backend(
    name: "Literal['lexv2models']",
) -> "BackendDict[LexModelsV2Backend]": ...
@overload
def get_backend(name: "Literal['logs']") -> "BackendDict[LogsBackend]": ...
@overload
def get_backend(
    name: "Literal['managedblockchain']",
) -> "BackendDict[ManagedBlockchainBackend]": ...
@overload
def get_backend(
    name: "Literal['mediaconnect']",
) -> "BackendDict[MediaConnectBackend]": ...
@overload
def get_backend(name: "Literal['medialive']") -> "BackendDict[MediaLiveBackend]": ...
@overload
def get_backend(
    name: "Literal['mediapackage']",
) -> "BackendDict[MediaPackageBackend]": ...
@overload
def get_backend(name: "Literal['mediastore']") -> "BackendDict[MediaStoreBackend]": ...
@overload
def get_backend(
    name: "Literal['mediastore-data']",
) -> "BackendDict[MediaStoreDataBackend]": ...
@overload
def get_backend(name: "Literal['memorydb']") -> "BackendDict[MemoryDBBackend]": ...
@overload
def get_backend(
    name: "Literal['meteringmarketplace']",
) -> "BackendDict[MeteringMarketplaceBackend]": ...
@overload
def get_backend(name: "Literal['moto_api']") -> "BackendDict[MotoAPIBackend]": ...
@overload
def get_backend(name: "Literal['mq']") -> "BackendDict[MQBackend]": ...
@overload
def get_backend(name: "Literal['neptune']") -> "BackendDict[RDSBackend]": ...
@overload
def get_backend(
    name: "Literal['networkmanager']",
) -> "BackendDict[NetworkManagerBackend]": ...
@overload
def get_backend(
    name: "Literal['opensearch']",
) -> "BackendDict[OpenSearchServiceBackend]": ...
@overload
def get_backend(
    name: "Literal['opensearchserverless']",
) -> "BackendDict[OpenSearchServiceServerlessBackend]": ...
@overload
def get_backend(name: "Literal['opsworks']") -> "BackendDict[OpsWorksBackend]": ...
@overload
def get_backend(
    name: "Literal['osis']",
) -> "BackendDict[OpenSearchIngestionBackend]": ...
@overload
def get_backend(
    name: "Literal['organizations']",
) -> "BackendDict[OrganizationsBackend]": ...
@overload
def get_backend(
    name: "Literal['personalize']",
) -> "BackendDict[PersonalizeBackend]": ...
@overload
def get_backend(name: "Literal['pinpoint']") -> "BackendDict[PinpointBackend]": ...
@overload
def get_backend(name: "Literal['polly']") -> "BackendDict[PollyBackend]": ...
@overload
def get_backend(name: "Literal['qldb']") -> "BackendDict[QLDBBackend]": ...
@overload
def get_backend(name: "Literal['quicksight']") -> "BackendDict[QuickSightBackend]": ...
@overload
def get_backend(
    name: "Literal['ram']",
) -> "BackendDict[ResourceAccessManagerBackend]": ...
@overload
def get_backend(name: "Literal['rds']") -> "BackendDict[RDSBackend]": ...
@overload
def get_backend(
    name: "Literal['rds-data']",
) -> "BackendDict[RDSDataServiceBackend]": ...
@overload
def get_backend(name: "Literal['redshift']") -> "BackendDict[RedshiftBackend]": ...
@overload
def get_backend(
    name: "Literal['redshift-data']",
) -> "BackendDict[RedshiftDataAPIServiceBackend]": ...
@overload
def get_backend(
    name: "Literal['rekognition']",
) -> "BackendDict[RekognitionBackend]": ...
@overload
def get_backend(
    name: "Literal['resiliencehub']",
) -> "BackendDict[ResilienceHubBackend]": ...
@overload
def get_backend(
    name: "Literal['resource-groups']",
) -> "BackendDict[ResourceGroupsBackend]": ...
@overload
def get_backend(
    name: "Literal['resourcegroupstaggingapi']",
) -> "BackendDict[ResourceGroupsTaggingAPIBackend]": ...
@overload
def get_backend(name: "Literal['robomaker']") -> "BackendDict[RoboMakerBackend]": ...
@overload
def get_backend(name: "Literal['route53']") -> "BackendDict[Route53Backend]": ...
@overload
def get_backend(
    name: "Literal['route53resolver']",
) -> "BackendDict[Route53ResolverBackend]": ...
@overload
def get_backend(
    name: "Literal['route53domains']",
) -> "BackendDict[Route53DomainsBackend]": ...
@overload
def get_backend(name: "Literal['s3']") -> "BackendDict[S3Backend]": ...
@overload
def get_backend(name: "Literal['s3bucket_path']") -> "BackendDict[S3Backend]": ...
@overload
def get_backend(name: "Literal['s3control']") -> "BackendDict[S3ControlBackend]": ...
@overload
def get_backend(
    name: "Literal['sagemaker']",
) -> "BackendDict[SageMakerModelBackend]": ...
@overload
def get_backend(
    name: "Literal['sagemaker-metrics']",
) -> "BackendDict[SageMakerMetricsBackend]": ...
@overload
def get_backend(
    name: "Literal['sagemaker-runtime']",
) -> "BackendDict[SageMakerRuntimeBackend]": ...
@overload
def get_backend(
    name: "Literal['scheduler']",
) -> "BackendDict[EventBridgeSchedulerBackend]": ...
@overload
def get_backend(name: "Literal['sdb']") -> "BackendDict[SimpleDBBackend]": ...
@overload
def get_backend(
    name: "Literal['secretsmanager']",
) -> "BackendDict[SecretsManagerBackend]": ...
@overload
def get_backend(
    name: "Literal['servicediscovery']",
) -> "BackendDict[ServiceDiscoveryBackend]": ...
@overload
def get_backend(
    name: "Literal['service-quotas']",
) -> "BackendDict[ServiceQuotasBackend]": ...
@overload
def get_backend(name: "Literal['ses']") -> "BackendDict[SESBackend]": ...
@overload
def get_backend(name: "Literal['sesv2']") -> "BackendDict[SESV2Backend]": ...
@overload
def get_backend(name: "Literal['shield']") -> "BackendDict[ShieldBackend]": ...
@overload
def get_backend(name: "Literal['signer']") -> "BackendDict[SignerBackend]": ...
@overload
def get_backend(name: "Literal['sns']") -> "BackendDict[SNSBackend]": ...
@overload
def get_backend(name: "Literal['sqs']") -> "BackendDict[SQSBackend]": ...
@overload
def get_backend(
    name: "Literal['ssm']",
) -> "BackendDict[SimpleSystemManagerBackend]": ...
@overload
def get_backend(name: "Literal['sso-admin']") -> "BackendDict[SSOAdminBackend]": ...
@overload
def get_backend(
    name: "Literal['stepfunctions']",
) -> "BackendDict[StepFunctionBackend]": ...
@overload
def get_backend(name: "Literal['sts']") -> "BackendDict[STSBackend]": ...
@overload
def get_backend(name: "Literal['support']") -> "BackendDict[SupportBackend]": ...
@overload
def get_backend(name: "Literal['swf']") -> "BackendDict[SWFBackend]": ...
@overload
def get_backend(name: "Literal['textract']") -> "BackendDict[TextractBackend]": ...
@overload
def get_backend(
    name: "Literal['timestream-influxdb']",
) -> "BackendDict[TimestreamInfluxDBBackend]": ...
@overload
def get_backend(
    name: "Literal['timestream-query']",
) -> "BackendDict[TimestreamQueryBackend]": ...
@overload
def get_backend(
    name: "Literal['timestream-write']",
) -> "BackendDict[TimestreamWriteBackend]": ...
@overload
def get_backend(name: "Literal['transcribe']") -> "BackendDict[TranscribeBackend]": ...
@overload
def get_backend(name: "Literal['transfer']") -> "BackendDict[TransferBackend]": ...
@overload
def get_backend(name: "Literal['wafv2']") -> "BackendDict[WAFV2Backend]": ...
@overload
def get_backend(name: "Literal['workspaces']") -> "BackendDict[WorkSpacesBackend]": ...
@overload
def get_backend(
    name: "Literal['workspaces-web']",
) -> "BackendDict[WorkSpacesWebBackend]": ...
@overload
def get_backend(name: "Literal['xray']") -> "BackendDict[XRayBackend]": ...
@overload
def get_backend(name: "Literal['s3tables']") -> "BackendDict[S3TablesBackend]": ...


def get_backend(name: SERVICE_NAMES) -> "BackendDict[SERVICE_BACKEND]":
    # fmt: on
    safe_name = name.replace("-", "")
    return _import_backend(
        ALT_SERVICE_NAMES.get(safe_name, safe_name),
        f"{ALT_BACKEND_NAMES.get(safe_name, safe_name)}_backends",
    )



from datetime import datetime
from uuid import uuid4

import boto3
import pytest
from botocore.exceptions import ClientError
from dateutil.tz import tzlocal

from moto import mock_aws
from moto.secretsmanager.list_secrets.filters import split_words
from tests import aws_verified


def boto_client():
    return boto3.client("secretsmanager", region_name="us-west-2")


@mock_aws
def test_empty():
    conn = boto_client()

    secrets = conn.list_secrets()

    assert secrets["SecretList"] == []


@mock_aws
def test_list_secrets():
    conn = boto_client()

    conn.create_secret(Name="test-secret", SecretString="foosecret")

    conn.create_secret(
        Name="test-secret-2",
        SecretString="barsecret",
        Tags=[{"Key": "a", "Value": "1"}],
    )

    secrets = conn.list_secrets()

    assert secrets["SecretList"][0]["ARN"] is not None
    assert secrets["SecretList"][0]["Name"] == "test-secret"
    assert secrets["SecretList"][0]["SecretVersionsToStages"] is not None
    assert secrets["SecretList"][1]["ARN"] is not None
    assert secrets["SecretList"][1]["Name"] == "test-secret-2"
    assert secrets["SecretList"][1]["Tags"] == [{"Key": "a", "Value": "1"}]
    assert secrets["SecretList"][1]["SecretVersionsToStages"] is not None
    assert secrets["SecretList"][0]["CreatedDate"] <= datetime.now(tz=tzlocal())
    assert secrets["SecretList"][1]["CreatedDate"] <= datetime.now(tz=tzlocal())
    assert secrets["SecretList"][0]["LastChangedDate"] <= datetime.now(tz=tzlocal())
    assert secrets["SecretList"][1]["LastChangedDate"] <= datetime.now(tz=tzlocal())


@mock_aws
def test_with_name_filter():
    conn = boto_client()

    conn.create_secret(Name="foo", SecretString="secret")
    conn.create_secret(Name="bar", SecretString="secret")

    secrets = conn.list_secrets(Filters=[{"Key": "name", "Values": ["foo"]}])

    secret_names = list(map(lambda s: s["Name"], secrets["SecretList"]))
    assert secret_names == ["foo"]


@mock_aws
def test_with_tag_key_filter():
    conn = boto_client()

    conn.create_secret(
        Name="foo", SecretString="secret", Tags=[{"Key": "baz", "Value": "1"}]
    )
    conn.create_secret(Name="bar", SecretString="secret")

    secrets = conn.list_secrets(Filters=[{"Key": "tag-key", "Values": ["baz"]}])

    secret_names = list(map(lambda s: s["Name"], secrets["SecretList"]))
    assert secret_names == ["foo"]


@mock_aws
def test_with_tag_value_filter():
    conn = boto_client()

    conn.create_secret(
        Name="foo", SecretString="secret", Tags=[{"Key": "1", "Value": "baz"}]
    )
    conn.create_secret(Name="bar", SecretString="secret")

    secrets = conn.list_secrets(Filters=[{"Key": "tag-value", "Values": ["baz"]}])

    secret_names = list(map(lambda s: s["Name"], secrets["SecretList"]))
    assert secret_names == ["foo"]


@mock_aws
def test_with_description_filter():
    conn = boto_client()

    conn.create_secret(Name="foo", SecretString="secret", Description="baz qux")
    conn.create_secret(Name="bar", SecretString="secret")

    secrets = conn.list_secrets(Filters=[{"Key": "description", "Values": ["baz"]}])

    secret_names = list(map(lambda s: s["Name"], secrets["SecretList"]))
    assert secret_names == ["foo"]


@aws_verified
# Verified, but not marked because it's flaky - AWS can take up to 5 minutes before secrets are listed
def test_with_all_filter():
    # The 'all' filter will match a secret that contains ANY field with
    # the criteria. In other words an implicit OR.
    unique_name = str(uuid4())[0:6]

    first_secret = f"SecretOne{unique_name}"
    second_secret = f"SecretTwo{unique_name}"
    third_secret = f"Thirdsecret{unique_name}"
    foo_tag_key = f"SecretFTag{unique_name}"
    foo_tag_val = f"SecretFValue{unique_name}"
    no_match = f"none{unique_name}"

    conn = boto_client()

    conn.create_secret(Name=first_secret, SecretString="s")
    conn.create_secret(Name=second_secret, SecretString="s", Description="DescTwo")
    conn.create_secret(Name=third_secret, SecretString="s")
    conn.create_secret(
        Name=foo_tag_key, SecretString="s", Tags=[{"Key": "1", "Value": "foo"}]
    )
    conn.create_secret(
        Name=foo_tag_val, SecretString="s", Tags=[{"Key": "foo", "Value": "v"}]
    )
    conn.create_secret(Name=no_match, SecretString="s")

    try:
        # Full Match
        secrets = conn.list_secrets(Filters=[{"Key": "all", "Values": ["SecretOne"]}])[
            "SecretList"
        ]
        secret_names = [s["Name"] for s in secrets]
        assert secret_names == [first_secret]

        # Search in tags
        secrets = conn.list_secrets(Filters=[{"Key": "all", "Values": ["foo"]}])[
            "SecretList"
        ]
        secret_names = [s["Name"] for s in secrets]
        assert secret_names == [foo_tag_key, foo_tag_val]

        # StartsWith - full word
        secrets = conn.list_secrets(Filters=[{"Key": "all", "Values": ["Secret"]}])[
            "SecretList"
        ]
        secret_names = [s["Name"] for s in secrets]
        assert secret_names == [first_secret, second_secret, foo_tag_key, foo_tag_val]

        # Partial Match - full word - lowercase
        secrets = conn.list_secrets(Filters=[{"Key": "all", "Values": ["secret"]}])[
            "SecretList"
        ]
        secret_names = [s["Name"] for s in secrets]
        assert secret_names == [first_secret, second_secret, foo_tag_key, foo_tag_val]

        # Partial Match - partial word
        secrets = conn.list_secrets(Filters=[{"Key": "all", "Values": ["ret"]}])[
            "SecretList"
        ]
        secret_names = [s["Name"] for s in secrets]
        assert not secret_names

        # Partial Match - partial word
        secrets = conn.list_secrets(Filters=[{"Key": "all", "Values": ["sec"]}])[
            "SecretList"
        ]
        secret_names = [s["Name"] for s in secrets]
        assert secret_names == [first_secret, second_secret, foo_tag_key, foo_tag_val]

        # Unknown Match
        secrets = conn.list_secrets(Filters=[{"Key": "all", "Values": ["SomeSecret"]}])[
            "SecretList"
        ]
        assert not secrets

        # Single value
        # Only matches description that contains a d
        secrets = conn.list_secrets(Filters=[{"Key": "all", "Values": ["d"]}])[
            "SecretList"
        ]
        secret_names = [s["Name"] for s in secrets]
        assert secret_names == [second_secret]

        # Multiple values
        # Only matches description that contains a d and t (DescTwo)
        secrets = conn.list_secrets(Filters=[{"Key": "all", "Values": ["d t"]}])[
            "SecretList"
        ]
        secret_names = [s["Name"] for s in secrets]
        assert secret_names == [second_secret]

        # Name parts that start with a t (DescTwo)
        secrets = conn.list_secrets(Filters=[{"Key": "all", "Values": ["t"]}])[
            "SecretList"
        ]
        secret_names = [s["Name"] for s in secrets]
        assert secret_names == [second_secret, third_secret, foo_tag_key]
    finally:
        conn.delete_secret(SecretId=first_secret, ForceDeleteWithoutRecovery=True)
        conn.delete_secret(SecretId=second_secret, ForceDeleteWithoutRecovery=True)
        conn.delete_secret(SecretId=third_secret, ForceDeleteWithoutRecovery=True)
        conn.delete_secret(SecretId=foo_tag_key, ForceDeleteWithoutRecovery=True)
        conn.delete_secret(SecretId=foo_tag_val, ForceDeleteWithoutRecovery=True)
        conn.delete_secret(SecretId=no_match, ForceDeleteWithoutRecovery=True)


@aws_verified
# Verified, but not marked because it's flaky - AWS can take up to 5 minutes before secrets are listed
def test_with_all_filter_special_characters():
    unique_name = str(uuid4())[0:6]
    secret_with_slash = f"prod/AppBeta/{unique_name}"
    secret_with_under = f"prod_AppBeta_{unique_name}"
    secret_with_plus = f"prod+AppBeta+{unique_name}"
    secret_with_equal = f"prod=AppBeta={unique_name}"
    secret_with_dot = f"prod.AppBeta.{unique_name}"
    secret_with_at = f"prod@AppBeta@{unique_name}"
    secret_with_dash = f"prod-AppBeta-{unique_name}"
    # Note that this secret is never found, because the pattern is unknown
    secret_with_dash_and_slash = f"prod-AppBeta/{unique_name}"
    full_uppercase = f"uat/COMPANY/{unique_name}"
    partial_uppercase = f"uat/COMPANYthings/{unique_name}"

    all_special_char_names = [
        secret_with_slash,
        secret_with_under,
        secret_with_plus,
        secret_with_equal,
        secret_with_dot,
        secret_with_at,
        secret_with_dash,
    ]

    conn = boto_client()

    conn.create_secret(Name=secret_with_slash, SecretString="s")
    conn.create_secret(Name=secret_with_under, SecretString="s")
    conn.create_secret(Name=secret_with_plus, SecretString="s")
    conn.create_secret(Name=secret_with_equal, SecretString="s")
    conn.create_secret(Name=secret_with_dot, SecretString="s")
    conn.create_secret(Name=secret_with_at, SecretString="s")
    conn.create_secret(Name=secret_with_dash, SecretString="s")
    conn.create_secret(Name=full_uppercase, SecretString="s")
    conn.create_secret(Name=partial_uppercase, SecretString="s")

    try:
        # Partial Match
        secrets = conn.list_secrets(Filters=[{"Key": "all", "Values": ["AppBeta"]}])
        secret_names = [s["Name"] for s in secrets["SecretList"]]
        assert secret_names == all_special_char_names

        # Partial Match
        secrets = conn.list_secrets(Filters=[{"Key": "all", "Values": ["Beta"]}])
        secret_names = [s["Name"] for s in secrets["SecretList"]]
        assert secret_names == all_special_char_names

        secrets = conn.list_secrets(
            Filters=[{"Key": "all", "Values": ["AppBeta", "prod"]}]
        )["SecretList"]
        secret_names = [s["Name"] for s in secrets]
        assert secret_names == all_special_char_names

        # Search for special character itself
        secrets = conn.list_secrets(Filters=[{"Key": "all", "Values": ["+"]}])
        secret_names = [s["Name"] for s in secrets["SecretList"]]
        assert not secret_names

        # Search for unique postfix
        secrets = conn.list_secrets(Filters=[{"Key": "all", "Values": [unique_name]}])
        secret_names = [s["Name"] for s in secrets["SecretList"]]
        assert secret_names == (
            all_special_char_names + [full_uppercase, partial_uppercase]
        )

        # Search for unique postfix
        secrets = conn.list_secrets(Filters=[{"Key": "all", "Values": ["company"]}])
        secret_names = [s["Name"] for s in secrets["SecretList"]]
        assert secret_names == [full_uppercase]

        # This on it's own is not a word
        secrets = conn.list_secrets(Filters=[{"Key": "all", "Values": ["things"]}])
        secret_names = [s["Name"] for s in secrets["SecretList"]]
        assert secret_names == []

        # This is valid, because it's split as COMPAN + Ythings
        secrets = conn.list_secrets(Filters=[{"Key": "all", "Values": ["Ythings"]}])
        secret_names = [s["Name"] for s in secrets["SecretList"]]
        assert secret_names == [partial_uppercase]

        # Note that individual letters from COMPANY are not searchable,
        # indicating that AWS splits by terms, rather than each individual upper case
        # COMPANYThings --> COMPAN, YThings
        secrets = conn.list_secrets(Filters=[{"Key": "all", "Values": ["N"]}])
        secret_names = [s["Name"] for s in secrets["SecretList"]]
        assert secret_names == []

        #
        secrets = conn.list_secrets(Filters=[{"Key": "all", "Values": ["pany"]}])
        secret_names = [s["Name"] for s in secrets["SecretList"]]
        assert secret_names == []

    finally:
        conn.delete_secret(SecretId=secret_with_slash, ForceDeleteWithoutRecovery=True)
        conn.delete_secret(SecretId=secret_with_under, ForceDeleteWithoutRecovery=True)
        conn.delete_secret(SecretId=secret_with_plus, ForceDeleteWithoutRecovery=True)
        conn.delete_secret(SecretId=secret_with_equal, ForceDeleteWithoutRecovery=True)
        conn.delete_secret(SecretId=secret_with_dot, ForceDeleteWithoutRecovery=True)
        conn.delete_secret(SecretId=secret_with_dash, ForceDeleteWithoutRecovery=True)
        conn.delete_secret(
            SecretId=secret_with_dash_and_slash, ForceDeleteWithoutRecovery=True
        )
        conn.delete_secret(SecretId=secret_with_at, ForceDeleteWithoutRecovery=True)
        conn.delete_secret(SecretId=full_uppercase, ForceDeleteWithoutRecovery=True)
        conn.delete_secret(SecretId=partial_uppercase, ForceDeleteWithoutRecovery=True)


@mock_aws
def test_with_no_filter_key():
    conn = boto_client()

    with pytest.raises(ClientError) as ire:
        conn.list_secrets(Filters=[{"Values": ["foo"]}])

    assert ire.value.response["Error"]["Code"] == "InvalidParameterException"
    assert ire.value.response["Error"]["Message"] == "Invalid filter key"


@mock_aws
def test_with_no_filter_values():
    conn = boto_client()

    conn.create_secret(Name="foo", SecretString="secret", Description="hello")

    with pytest.raises(ClientError) as ire:
        conn.list_secrets(Filters=[{"Key": "description"}])

    assert ire.value.response["Error"]["Code"] == "InvalidParameterException"
    assert ire.value.response["Error"]["Message"] == (
        "Invalid filter values for key: description"
    )


@mock_aws
def test_with_invalid_filter_key():
    conn = boto_client()

    with pytest.raises(ClientError) as ire:
        conn.list_secrets(Filters=[{"Key": "invalid", "Values": ["foo"]}])

    assert ire.value.response["Error"]["Code"] == "ValidationException"
    assert ire.value.response["Error"]["Message"] == (
        "1 validation error detected: Value 'invalid' at "
        "'filters.1.member.key' failed to satisfy constraint: Member "
        "must satisfy enum value set: [all, name, tag-key, description, tag-value]"
    )


@aws_verified
# Verified, but not marked because it's flaky - AWS can take up to 5 minutes before secrets are listed
def test_with_duplicate_filter_keys():
    # Multiple filters with the same key combine with an implicit AND operator

    conn = boto_client()

    conn.create_secret(Name="foo", SecretString="secret", Description="one two")
    conn.create_secret(Name="bar", SecretString="secret", Description="one")
    conn.create_secret(Name="baz", SecretString="secret", Description="two")
    conn.create_secret(Name="qux", SecretString="secret", Description="unrelated")

    try:
        secrets = conn.list_secrets(
            Filters=[
                {"Key": "description", "Values": ["one"]},
                {"Key": "description", "Values": ["two"]},
            ]
        )

        secret_names = list(map(lambda s: s["Name"], secrets["SecretList"]))
        assert secret_names == ["foo"]
    finally:
        conn.delete_secret(SecretId="foo", ForceDeleteWithoutRecovery=True)
        conn.delete_secret(SecretId="bar", ForceDeleteWithoutRecovery=True)
        conn.delete_secret(SecretId="baz", ForceDeleteWithoutRecovery=True)
        conn.delete_secret(SecretId="qux", ForceDeleteWithoutRecovery=True)


@mock_aws
def test_with_multiple_filters():
    # Multiple filters combine with an implicit AND operator

    conn = boto_client()

    conn.create_secret(
        Name="foo", SecretString="secret", Tags=[{"Key": "right", "Value": "right"}]
    )
    conn.create_secret(
        Name="bar", SecretString="secret", Tags=[{"Key": "right", "Value": "wrong"}]
    )
    conn.create_secret(
        Name="baz", SecretString="secret", Tags=[{"Key": "wrong", "Value": "right"}]
    )
    conn.create_secret(
        Name="qux", SecretString="secret", Tags=[{"Key": "wrong", "Value": "wrong"}]
    )

    secrets = conn.list_secrets(
        Filters=[
            {"Key": "tag-key", "Values": ["right"]},
            {"Key": "tag-value", "Values": ["right"]},
        ]
    )

    secret_names = list(map(lambda s: s["Name"], secrets["SecretList"]))
    assert secret_names == ["foo"]


@mock_aws
def test_with_filter_with_multiple_values():
    conn = boto_client()

    conn.create_secret(Name="foo", SecretString="secret")
    conn.create_secret(Name="bar", SecretString="secret")
    conn.create_secret(Name="baz", SecretString="secret")

    secrets = conn.list_secrets(Filters=[{"Key": "name", "Values": ["foo", "bar"]}])

    secret_names = list(map(lambda s: s["Name"], secrets["SecretList"]))
    assert secret_names == ["foo", "bar"]


@aws_verified
# Verified, but not marked because it's flaky - AWS can take up to 5 minutes before secrets are listed
def test_with_filter_with_value_with_multiple_words():
    conn = boto_client()

    conn.create_secret(Name="foo", SecretString="secret", Description="one two")
    conn.create_secret(Name="bar", SecretString="secret", Description="one and two")
    conn.create_secret(Name="baz", SecretString="secret", Description="one")
    conn.create_secret(Name="qux", SecretString="secret", Description="two")
    conn.create_secret(Name="none", SecretString="secret", Description="unrelated")

    try:
        # All values that contain one and two
        secrets = conn.list_secrets(
            Filters=[{"Key": "description", "Values": ["one two"]}]
        )
        secret_names = list(map(lambda s: s["Name"], secrets["SecretList"]))
        assert secret_names == ["foo", "bar"]

        # All values that start with o and t
        secrets = conn.list_secrets(Filters=[{"Key": "description", "Values": ["o t"]}])
        secret_names = list(map(lambda s: s["Name"], secrets["SecretList"]))
        assert secret_names == ["foo", "bar"]

        # All values that contain t
        secrets = conn.list_secrets(Filters=[{"Key": "description", "Values": ["t"]}])
        secret_names = list(map(lambda s: s["Name"], secrets["SecretList"]))
        assert secret_names == ["foo", "bar", "qux"]
    finally:
        conn.delete_secret(SecretId="foo", ForceDeleteWithoutRecovery=True)
        conn.delete_secret(SecretId="bar", ForceDeleteWithoutRecovery=True)
        conn.delete_secret(SecretId="baz", ForceDeleteWithoutRecovery=True)
        conn.delete_secret(SecretId="qux", ForceDeleteWithoutRecovery=True)
        conn.delete_secret(SecretId="none", ForceDeleteWithoutRecovery=True)


@mock_aws
def test_with_filter_with_negation():
    conn = boto_client()

    conn.create_secret(Name="foo", SecretString="secret", Description="one two")
    conn.create_secret(Name="bar", SecretString="secret", Description="one and two")
    conn.create_secret(Name="baz", SecretString="secret", Description="one")
    conn.create_secret(Name="qux", SecretString="secret", Description="two")
    conn.create_secret(Name="none", SecretString="secret", Description="unrelated")

    secrets = conn.list_secrets(
        Filters=[{"Key": "description", "Values": ["one", "!two"]}]
    )

    secret_names = list(map(lambda s: s["Name"], secrets["SecretList"]))
    for secret_name in ["foo", "bar", "baz"]:
        assert secret_name in secret_names

    secrets = conn.list_secrets(Filters=[{"Key": "description", "Values": ["!o"]}])
    secret_names = list(map(lambda s: s["Name"], secrets["SecretList"]))
    for secret_name in ["qux", "none"]:
        assert secret_name in secret_names


@mock_aws
def test_filter_with_owning_service():
    conn = boto3.client("secretsmanager", region_name="us-west-2")

    conn.create_secret(Name="foo", SecretString="secret")

    resp = conn.list_secrets(Filters=[{"Key": "owning-service", "Values": ["n/a"]}])
    assert resp["SecretList"] == []


@mock_aws
def test_with_include_planned_deleted_secrets():
    conn = boto_client()

    conn.create_secret(Name="foo", SecretString="secret")
    conn.create_secret(Name="bar", SecretString="secret")

    secrets = conn.list_secrets()
    assert len(secrets["SecretList"]) == 2

    conn.delete_secret(SecretId="foo")

    # By default list secrets doesn't include deleted secrets
    secrets = conn.list_secrets()
    assert len(secrets["SecretList"]) == 1
    assert secrets["SecretList"][0]["ARN"] is not None
    assert secrets["SecretList"][0]["Name"] == "bar"
    assert secrets["SecretList"][0]["SecretVersionsToStages"] is not None

    # list secrets when IncludePlannedDeletion param included
    secrets = conn.list_secrets(IncludePlannedDeletion=True)
    assert len(secrets["SecretList"]) == 2

    # list secret with filter and IncludePlannedDeletion params
    secrets = conn.list_secrets(
        IncludePlannedDeletion=True, Filters=[{"Key": "name", "Values": ["foo"]}]
    )
    assert len(secrets["SecretList"]) == 1
    assert secrets["SecretList"][0]["ARN"] is not None
    assert secrets["SecretList"][0]["Name"] == "foo"
    assert secrets["SecretList"][0]["SecretVersionsToStages"] is not None


@pytest.mark.parametrize(
    "input,output",
    [
        ("test", ["test"]),
        ("my test", ["my", "test"]),
        ("Mytest", ["Mytest"]),
        ("MyTest", ["My", "Test"]),
        ("MyTestPhrase", ["My", "Test", "Phrase"]),
        ("myTest", ["my", "Test"]),
        ("my test", ["my", "test"]),
        ("my/test", ["my", "test"]),
    ],
)
def test_word_splitter(input, output):
    assert split_words(input) == output



import copy
from typing import Final

from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.jsonata.jsonata import (
    VariableDeclarations,
    compose_jsonata_expression,
    eval_jsonata_expression,
)
from moto.stepfunctions.parser.asl.utils.json_path import extract_json


class ItemsPath(EvalComponent):
    DEFAULT_PATH: Final[str] = "$"

    def __init__(self, path: str = DEFAULT_PATH):
        self.path = path

    def _eval_body(self, env: Environment) -> None:
        value = copy.deepcopy(env.stack[-1])
        if self.path != ItemsPath.DEFAULT_PATH:
            value = extract_json(self.path, value)
        env.stack.append(value)


class ItemsPathContextObject(ItemsPath):
    def __init__(self, path: str):
        path_tail = path[1:]
        super().__init__(path=path_tail)

    def _eval_body(self, env: Environment) -> None:
        value = extract_json(self.path, env.states.context_object.context_object_data)
        env.stack.append(copy.deepcopy(value))


class ItemsPathVar(ItemsPath):
    def _eval_body(self, env: Environment) -> None:
        variable_declarations: VariableDeclarations = (
            env.variable_store.get_variable_declarations()
        )
        jsonata_expression = compose_jsonata_expression(
            final_jsonata_expression=self.path,  # noqa
            variable_declarations_list=[variable_declarations],
        )
        value = eval_jsonata_expression(jsonata_expression=jsonata_expression)
        env.stack.append(copy.deepcopy(value))



"""Handles incoming memorydb requests, invokes methods, returns responses."""

import json

from moto.core.responses import BaseResponse

from .models import MemoryDBBackend, memorydb_backends


class MemoryDBResponse(BaseResponse):
    """Handler for MemoryDB requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="memorydb")

    @property
    def memorydb_backend(self) -> MemoryDBBackend:
        """Return backend instance specific for this region."""
        return memorydb_backends[self.current_account][self.region]

    def create_cluster(self) -> str:
        params = json.loads(self.body)
        cluster_name = params.get("ClusterName")
        node_type = params.get("NodeType")
        parameter_group_name = params.get("ParameterGroupName")
        description = params.get("Description")
        num_shards = params.get("NumShards")
        num_replicas_per_shard = params.get("NumReplicasPerShard")
        subnet_group_name = params.get("SubnetGroupName")
        security_group_ids = params.get("SecurityGroupIds")
        maintenance_window = params.get("MaintenanceWindow")
        port = params.get("Port")
        sns_topic_arn = params.get("SnsTopicArn")
        tls_enabled = params.get("TLSEnabled")
        kms_key_id = params.get("KmsKeyId")
        snapshot_arns = params.get("SnapshotArns")
        snapshot_name = params.get("SnapshotName")
        snapshot_retention_limit = params.get("SnapshotRetentionLimit")
        tags = params.get("Tags")
        snapshot_window = params.get("SnapshotWindow")
        acl_name = params.get("ACLName")
        engine_version = params.get("EngineVersion")
        auto_minor_version_upgrade = params.get("AutoMinorVersionUpgrade")
        data_tiering = params.get("DataTiering")
        cluster = self.memorydb_backend.create_cluster(
            cluster_name=cluster_name,
            node_type=node_type,
            parameter_group_name=parameter_group_name,
            description=description,
            num_shards=num_shards,
            num_replicas_per_shard=num_replicas_per_shard,
            subnet_group_name=subnet_group_name,
            security_group_ids=security_group_ids,
            maintenance_window=maintenance_window,
            port=port,
            sns_topic_arn=sns_topic_arn,
            tls_enabled=tls_enabled,
            kms_key_id=kms_key_id,
            snapshot_arns=snapshot_arns,
            snapshot_name=snapshot_name,
            snapshot_retention_limit=snapshot_retention_limit,
            tags=tags,
            snapshot_window=snapshot_window,
            acl_name=acl_name,
            engine_version=engine_version,
            auto_minor_version_upgrade=auto_minor_version_upgrade,
            data_tiering=data_tiering,
        )
        return json.dumps(dict(Cluster=cluster.to_dict()))

    def create_subnet_group(self) -> str:
        params = json.loads(self.body)
        subnet_group_name = params.get("SubnetGroupName")
        description = params.get("Description")
        subnet_ids = params.get("SubnetIds")
        tags = params.get("Tags")
        subnet_group = self.memorydb_backend.create_subnet_group(
            subnet_group_name=subnet_group_name,
            description=description,
            subnet_ids=subnet_ids,
            tags=tags,
        )
        return json.dumps(dict(SubnetGroup=subnet_group.to_dict()))

    def create_snapshot(self) -> str:
        params = json.loads(self.body)
        cluster_name = params.get("ClusterName")
        snapshot_name = params.get("SnapshotName")
        kms_key_id = params.get("KmsKeyId")
        tags = params.get("Tags")
        snapshot = self.memorydb_backend.create_snapshot(
            cluster_name=cluster_name,
            snapshot_name=snapshot_name,
            kms_key_id=kms_key_id,
            tags=tags,
        )
        return json.dumps(dict(Snapshot=snapshot.to_dict()))

    def describe_clusters(self) -> str:
        params = json.loads(self.body)
        cluster_name = params.get("ClusterName")
        show_shard_details = params.get("ShowShardDetails")
        clusters = self.memorydb_backend.describe_clusters(
            cluster_name=cluster_name,
        )
        return json.dumps(
            dict(
                Clusters=[
                    cluster.to_desc_dict() if show_shard_details else cluster.to_dict()
                    for cluster in clusters
                ]
            )
        )

    def describe_snapshots(self) -> str:
        params = json.loads(self.body)
        cluster_name = params.get("ClusterName")
        snapshot_name = params.get("SnapshotName")
        source = params.get("Source")
        show_detail = params.get("ShowDetail")
        snapshots = self.memorydb_backend.describe_snapshots(
            cluster_name=cluster_name,
            snapshot_name=snapshot_name,
            source=source,
        )
        return json.dumps(
            dict(
                Snapshots=[
                    snapshot.to_desc_dict() if show_detail else snapshot.to_dict()
                    for snapshot in snapshots
                ]
            )
        )

    def describe_subnet_groups(self) -> str:
        params = json.loads(self.body)
        subnet_group_name = params.get("SubnetGroupName")
        subnet_groups = self.memorydb_backend.describe_subnet_groups(
            subnet_group_name=subnet_group_name,
        )
        return json.dumps(dict(SubnetGroups=[sg.to_dict() for sg in subnet_groups]))

    def list_tags(self) -> str:
        params = json.loads(self.body)
        resource_arn = params.get("ResourceArn")
        tag_list = self.memorydb_backend.list_tags(
            resource_arn=resource_arn,
        )
        return json.dumps(dict(TagList=tag_list))

    def tag_resource(self) -> str:
        params = json.loads(self.body)
        resource_arn = params.get("ResourceArn")
        tags = params.get("Tags")
        tag_list = self.memorydb_backend.tag_resource(
            resource_arn=resource_arn,
            tags=tags,
        )
        return json.dumps(dict(TagList=tag_list))

    def untag_resource(self) -> str:
        params = json.loads(self.body)
        resource_arn = params.get("ResourceArn")
        tag_keys = params.get("TagKeys")
        tag_list = self.memorydb_backend.untag_resource(
            resource_arn=resource_arn,
            tag_keys=tag_keys,
        )
        return json.dumps(dict(TagList=tag_list))

    def update_cluster(self) -> str:
        params = json.loads(self.body)
        cluster_name = params.get("ClusterName")
        description = params.get("Description")
        security_group_ids = params.get("SecurityGroupIds")
        maintenance_window = params.get("MaintenanceWindow")
        sns_topic_arn = params.get("SnsTopicArn")
        sns_topic_status = params.get("SnsTopicStatus")
        parameter_group_name = params.get("ParameterGroupName")
        snapshot_window = params.get("SnapshotWindow")
        snapshot_retention_limit = params.get("SnapshotRetentionLimit")
        node_type = params.get("NodeType")
        engine_version = params.get("EngineVersion")
        replica_configuration = params.get("ReplicaConfiguration")
        shard_configuration = params.get("ShardConfiguration")
        acl_name = params.get("ACLName")
        cluster = self.memorydb_backend.update_cluster(
            cluster_name=cluster_name,
            description=description,
            security_group_ids=security_group_ids,
            maintenance_window=maintenance_window,
            sns_topic_arn=sns_topic_arn,
            sns_topic_status=sns_topic_status,
            parameter_group_name=parameter_group_name,
            snapshot_window=snapshot_window,
            snapshot_retention_limit=snapshot_retention_limit,
            node_type=node_type,
            engine_version=engine_version,
            replica_configuration=replica_configuration,
            shard_configuration=shard_configuration,
            acl_name=acl_name,
        )
        return json.dumps(dict(Cluster=cluster.to_dict()))

    def delete_cluster(self) -> str:
        params = json.loads(self.body)
        cluster_name = params.get("ClusterName")
        final_snapshot_name = params.get("FinalSnapshotName")
        cluster = self.memorydb_backend.delete_cluster(
            cluster_name=cluster_name,
            final_snapshot_name=final_snapshot_name,
        )
        return json.dumps(dict(Cluster=cluster.to_dict()))

    def delete_snapshot(self) -> str:
        params = json.loads(self.body)
        snapshot_name = params.get("SnapshotName")
        snapshot = self.memorydb_backend.delete_snapshot(
            snapshot_name=snapshot_name,
        )
        return json.dumps(dict(Snapshot=snapshot.to_dict()))

    def delete_subnet_group(self) -> str:
        params = json.loads(self.body)
        subnet_group_name = params.get("SubnetGroupName")
        subnet_group = self.memorydb_backend.delete_subnet_group(
            subnet_group_name=subnet_group_name,
        )
        return json.dumps(dict(SubnetGroup=subnet_group.to_dict()))



from __future__ import annotations

from typing import Any, Dict, Optional, Set

from moto.stepfunctions.parser.asl.jsonata.jsonata import (
    VariableDeclarations,
    encode_jsonata_variable_declarations,
)
from moto.stepfunctions.parser.asl.utils.encoding import to_json_str

VariableIdentifier = str
VariableValue = Any


class VariableStoreError(RuntimeError):
    message: str

    def __init__(self, message: str):
        self.message = message

    def __str__(self):
        return f"{self.__class__.__name__} {self.message}"

    def __repr__(self):
        return str(self)


class NoSuchVariable(VariableStoreError):
    variable_identifier: VariableIdentifier

    def __init__(self, variable_identifier: VariableIdentifier):
        super().__init__(message=f"No such variable '{variable_identifier}' in scope")
        self.variable_identifier = variable_identifier


class IllegalOuterScopeWrite(VariableStoreError):
    variable_identifier: VariableIdentifier
    variable_value: VariableValue

    def __init__(
        self, variable_identifier: VariableIdentifier, variable_value: VariableValue
    ):
        super().__init__(
            message=f"Cannot bind value '{variable_value}' to variable '{variable_identifier}' as it belongs to an outer scope."
        )
        self.variable_identifier = variable_identifier
        self.variable_value = variable_value


class VariableStore:
    _outer_scope: dict
    _inner_scope: dict

    _declaration_tracing: Set[str]

    _outer_variable_declaration_cache: Optional[VariableDeclarations]
    _variable_declarations_cache: Optional[VariableDeclarations]

    def __init__(self):
        self._outer_scope = dict()
        self._inner_scope = dict()
        self._declaration_tracing = set()
        self._outer_variable_declaration_cache = None
        self._variable_declarations_cache = None

    @classmethod
    def as_inner_scope_of(cls, outer_variable_store: VariableStore) -> VariableStore:
        inner_variable_store = cls()
        inner_variable_store._outer_scope.update(outer_variable_store._outer_scope)
        inner_variable_store._outer_scope.update(outer_variable_store._inner_scope)
        return inner_variable_store

    def reset_tracing(self) -> None:
        self._declaration_tracing.clear()

    # TODO: add typing when this available in service init.
    def get_assigned_variables(self) -> Dict[str, str]:
        assigned_variables: Dict[str, str] = dict()
        for traced_declaration_identifier in self._declaration_tracing:
            traced_declaration_value = self.get(traced_declaration_identifier)
            if isinstance(traced_declaration_value, str):
                traced_declaration_value_json_str = f'"{traced_declaration_value}"'
            else:
                traced_declaration_value_json_str: str = to_json_str(
                    traced_declaration_value, separators=(",", ":")
                )
            assigned_variables[traced_declaration_identifier] = (
                traced_declaration_value_json_str
            )
        return assigned_variables

    def get(self, variable_identifier: VariableIdentifier) -> VariableValue:
        if variable_identifier in self._inner_scope:
            return self._inner_scope[variable_identifier]
        if variable_identifier in self._outer_scope:
            return self._outer_scope[variable_identifier]
        raise NoSuchVariable(variable_identifier=variable_identifier)

    def set(
        self, variable_identifier: VariableIdentifier, variable_value: VariableValue
    ) -> None:
        if variable_identifier in self._outer_scope:
            raise IllegalOuterScopeWrite(
                variable_identifier=variable_identifier, variable_value=variable_value
            )
        self._declaration_tracing.add(variable_identifier)
        self._inner_scope[variable_identifier] = variable_value
        self._variable_declarations_cache = None

    @staticmethod
    def _to_variable_declarations(bindings: Dict[str, Any]) -> VariableDeclarations:
        variables = {f"${key}": value for key, value in bindings.items()}
        encoded = encode_jsonata_variable_declarations(variables)
        return encoded

    def get_variable_declarations(self) -> VariableDeclarations:
        if self._variable_declarations_cache is not None:
            return self._variable_declarations_cache
        if self._outer_variable_declaration_cache is None:
            self._outer_variable_declaration_cache = self._to_variable_declarations(
                self._outer_scope
            )
        inner_variable_declarations_cache = self._to_variable_declarations(
            self._inner_scope
        )
        self._variable_declarations_cache = "".join(
            [self._outer_variable_declaration_cache, inner_variable_declarations_cache]
        )
        return self._variable_declarations_cache



"""Unit tests for sagemaker-supported APIs."""

import datetime

import boto3

from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@mock_aws
def test_create_model_explainability_job_definition():
    client = boto3.client("sagemaker", region_name="eu-west-1")
    arn = client.create_model_explainability_job_definition(
        JobDefinitionName="testmodelexplainabilityjobdefinition",
        ModelExplainabilityBaselineConfig={
            "BaseliningJobName": "testbaseliningjobname",
            "ConstraintsResource": {"S3Uri": "s3://constrain/resource"},
        },
        ModelExplainabilityAppSpecification={
            "ImageUri": "testimageuri",
            "ConfigUri": "testconfiguri",
            "Environment": {"test": "test"},
        },
        ModelExplainabilityJobInput={
            "EndpointInput": {
                "EndpointName": "testendpointname",
                "LocalPath": "testlocalpath",
                "S3DataDistributionType": "FullyReplicated",
                "S3InputMode": "File",
            },
        },
        ModelExplainabilityJobOutputConfig={
            "MonitoringOutputs": [
                {
                    "S3Output": {
                        "S3Uri": "s3://output/uri",
                        "LocalPath": "testlocalpath",
                        "S3UploadMode": "Continuous",
                    }
                },
            ]
        },
        JobResources={
            "ClusterConfig": {
                "InstanceCount": 1,
                "InstanceType": "ml.m4.xlarge",
                "VolumeSizeInGB": 4,
            },
        },
        NetworkConfig={"EnableInterContainerTrafficEncryption": True},
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole",
        StoppingCondition={"MaxRuntimeInSeconds": 123},
        Tags=[{"Key": "testkey", "Value": "testvalue"}],
    )
    assert (
        arn["JobDefinitionArn"]
        == f"arn:aws:sagemaker:eu-west-1:{ACCOUNT_ID}:model-explainability-job-definition/testmodelexplainabilityjobdefinition"
    )


@mock_aws
def test_describe_model_explainability_job_definition():
    client = boto3.client("sagemaker", region_name="eu-west-1")
    client.create_model_explainability_job_definition(
        JobDefinitionName="testmodelexplainabilityjobdefinition",
        ModelExplainabilityBaselineConfig={
            "BaseliningJobName": "testbaseliningjobname",
            "ConstraintsResource": {"S3Uri": "s3://constrain/resource"},
        },
        ModelExplainabilityAppSpecification={
            "ImageUri": "testimageuri",
            "ConfigUri": "testconfiguri",
            "Environment": {"test": "test"},
        },
        ModelExplainabilityJobInput={
            "EndpointInput": {
                "EndpointName": "testendpointname",
                "LocalPath": "testlocalpath",
                "S3DataDistributionType": "FullyReplicated",
                "S3InputMode": "File",
            },
        },
        ModelExplainabilityJobOutputConfig={
            "MonitoringOutputs": [
                {
                    "S3Output": {
                        "S3Uri": "s3://output/uri",
                        "LocalPath": "testlocalpath",
                        "S3UploadMode": "Continuous",
                    }
                },
            ]
        },
        JobResources={
            "ClusterConfig": {
                "InstanceCount": 1,
                "InstanceType": "ml.m4.xlarge",
                "VolumeSizeInGB": 4,
            },
        },
        NetworkConfig={"EnableInterContainerTrafficEncryption": True},
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole",
        StoppingCondition={"MaxRuntimeInSeconds": 123},
        Tags=[{"Key": "testkey", "Value": "testvalue"}],
    )
    resp = client.describe_model_explainability_job_definition(
        JobDefinitionName="testmodelexplainabilityjobdefinition"
    )
    assert (
        resp["JobDefinitionArn"]
        == f"arn:aws:sagemaker:eu-west-1:{ACCOUNT_ID}:model-explainability-job-definition/testmodelexplainabilityjobdefinition"
    )
    assert resp["JobDefinitionName"] == "testmodelexplainabilityjobdefinition"
    assert resp["ModelExplainabilityBaselineConfig"] == {
        "BaseliningJobName": "testbaseliningjobname",
        "ConstraintsResource": {"S3Uri": "s3://constrain/resource"},
    }
    assert resp["ModelExplainabilityAppSpecification"] == {
        "ImageUri": "testimageuri",
        "ConfigUri": "testconfiguri",
        "Environment": {"test": "test"},
    }
    assert resp["ModelExplainabilityJobInput"] == {
        "EndpointInput": {
            "EndpointName": "testendpointname",
            "LocalPath": "testlocalpath",
            "S3DataDistributionType": "FullyReplicated",
            "S3InputMode": "File",
        },
    }
    assert resp["ModelExplainabilityJobOutputConfig"] == {
        "MonitoringOutputs": [
            {
                "S3Output": {
                    "S3Uri": "s3://output/uri",
                    "LocalPath": "testlocalpath",
                    "S3UploadMode": "Continuous",
                }
            },
        ]
    }
    assert resp["JobResources"] == {
        "ClusterConfig": {
            "InstanceCount": 1,
            "InstanceType": "ml.m4.xlarge",
            "VolumeSizeInGB": 4,
        },
    }
    assert resp["NetworkConfig"] == {"EnableInterContainerTrafficEncryption": True}
    assert resp["RoleArn"] == f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole"


@mock_aws
def test_describe_model_explainability_job_definition_defaults():
    client = boto3.client("sagemaker", region_name="eu-west-1")
    client.create_model_explainability_job_definition(
        JobDefinitionName="testmodelexplainabilityjobdefinition",
        ModelExplainabilityAppSpecification={
            "ImageUri": "testimageuri",
            "ConfigUri": "testconfiguri",
        },
        ModelExplainabilityJobInput={
            "EndpointInput": {
                "EndpointName": "testendpointname",
                "LocalPath": "testlocalpath",
            },
        },
        ModelExplainabilityJobOutputConfig={
            "MonitoringOutputs": [
                {
                    "S3Output": {
                        "S3Uri": "s3://output/uri",
                        "LocalPath": "testlocalpath",
                    }
                },
            ]
        },
        JobResources={
            "ClusterConfig": {
                "InstanceCount": 1,
                "InstanceType": "ml.m4.xlarge",
                "VolumeSizeInGB": 4,
            },
        },
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole",
    )
    resp = client.describe_model_explainability_job_definition(
        JobDefinitionName="testmodelexplainabilityjobdefinition"
    )
    assert (
        resp["JobDefinitionArn"]
        == f"arn:aws:sagemaker:eu-west-1:{ACCOUNT_ID}:model-explainability-job-definition/testmodelexplainabilityjobdefinition"
    )
    assert resp["JobDefinitionName"] == "testmodelexplainabilityjobdefinition"
    assert resp["ModelExplainabilityAppSpecification"] == {
        "ImageUri": "testimageuri",
        "ConfigUri": "testconfiguri",
    }
    assert resp["ModelExplainabilityJobInput"] == {
        "EndpointInput": {
            "EndpointName": "testendpointname",
            "LocalPath": "testlocalpath",
        },
    }
    assert resp["ModelExplainabilityJobOutputConfig"] == {
        "MonitoringOutputs": [
            {
                "S3Output": {
                    "S3Uri": "s3://output/uri",
                    "LocalPath": "testlocalpath",
                }
            },
        ]
    }
    assert resp["JobResources"] == {
        "ClusterConfig": {
            "InstanceCount": 1,
            "InstanceType": "ml.m4.xlarge",
            "VolumeSizeInGB": 4,
        },
    }
    assert resp["RoleArn"] == f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole"


@mock_aws
def test_list_model_explainability_job_definitions():
    client = boto3.client("sagemaker", region_name="eu-west-1")
    client.create_model_explainability_job_definition(
        JobDefinitionName="testmodelexplainabilityjobdefinition",
        ModelExplainabilityAppSpecification={
            "ImageUri": "testimageuri",
            "ConfigUri": "testconfiguri",
        },
        ModelExplainabilityJobInput={
            "EndpointInput": {
                "EndpointName": "testendpointname",
                "LocalPath": "testlocalpath",
            },
        },
        ModelExplainabilityJobOutputConfig={
            "MonitoringOutputs": [
                {
                    "S3Output": {
                        "S3Uri": "s3://output/uri",
                        "LocalPath": "testlocalpath",
                    }
                },
            ]
        },
        JobResources={
            "ClusterConfig": {
                "InstanceCount": 1,
                "InstanceType": "ml.m4.xlarge",
                "VolumeSizeInGB": 4,
            },
        },
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole",
    )
    client.create_model_explainability_job_definition(
        JobDefinitionName="testmodelexplainabilityjobdefinition2",
        ModelExplainabilityAppSpecification={
            "ImageUri": "testimageuri",
            "ConfigUri": "testconfiguri",
        },
        ModelExplainabilityJobInput={
            "EndpointInput": {
                "EndpointName": "testendpointname2",
                "LocalPath": "testlocalpath",
            },
        },
        ModelExplainabilityJobOutputConfig={
            "MonitoringOutputs": [
                {
                    "S3Output": {
                        "S3Uri": "s3://output/uri",
                        "LocalPath": "testlocalpath",
                    }
                },
            ]
        },
        JobResources={
            "ClusterConfig": {
                "InstanceCount": 1,
                "InstanceType": "ml.m4.xlarge",
                "VolumeSizeInGB": 4,
            },
        },
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole",
    )
    resp = client.list_model_explainability_job_definitions()["JobDefinitionSummaries"]
    assert (
        resp[0]["MonitoringJobDefinitionName"] == "testmodelexplainabilityjobdefinition"
    )
    assert (
        resp[0]["MonitoringJobDefinitionArn"]
        == f"arn:aws:sagemaker:eu-west-1:{ACCOUNT_ID}:model-explainability-job-definition/testmodelexplainabilityjobdefinition"
    )
    assert isinstance(resp[0]["CreationTime"], datetime.datetime)
    assert resp[0]["EndpointName"] == "testendpointname"
    assert (
        resp[1]["MonitoringJobDefinitionName"]
        == "testmodelexplainabilityjobdefinition2"
    )
    assert (
        resp[1]["MonitoringJobDefinitionArn"]
        == f"arn:aws:sagemaker:eu-west-1:{ACCOUNT_ID}:model-explainability-job-definition/testmodelexplainabilityjobdefinition2"
    )
    assert isinstance(resp[1]["CreationTime"], datetime.datetime)
    assert resp[1]["EndpointName"] == "testendpointname2"


@mock_aws
def test_list_model_explainability_job_definitions_filters():
    client = boto3.client("sagemaker", region_name="eu-west-1")
    client.create_model_explainability_job_definition(
        JobDefinitionName="testmodelexplainabilityjobdefinition",
        ModelExplainabilityAppSpecification={
            "ImageUri": "testimageuri",
            "ConfigUri": "testconfiguri",
        },
        ModelExplainabilityJobInput={
            "EndpointInput": {
                "EndpointName": "testendpointname",
                "LocalPath": "testlocalpath",
            },
        },
        ModelExplainabilityJobOutputConfig={
            "MonitoringOutputs": [
                {
                    "S3Output": {
                        "S3Uri": "s3://output/uri",
                        "LocalPath": "testlocalpath",
                    }
                },
            ]
        },
        JobResources={
            "ClusterConfig": {
                "InstanceCount": 1,
                "InstanceType": "ml.m4.xlarge",
                "VolumeSizeInGB": 4,
            },
        },
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole",
    )
    client.create_model_explainability_job_definition(
        JobDefinitionName="test2modelexplainabilityjobdefinition2",
        ModelExplainabilityAppSpecification={
            "ImageUri": "testimageuri",
            "ConfigUri": "testconfiguri",
        },
        ModelExplainabilityJobInput={
            "EndpointInput": {
                "EndpointName": "testendpointname",
                "LocalPath": "testlocalpath",
            },
        },
        ModelExplainabilityJobOutputConfig={
            "MonitoringOutputs": [
                {
                    "S3Output": {
                        "S3Uri": "s3://output/uri",
                        "LocalPath": "testlocalpath",
                    }
                },
            ]
        },
        JobResources={
            "ClusterConfig": {
                "InstanceCount": 1,
                "InstanceType": "ml.m4.xlarge",
                "VolumeSizeInGB": 4,
            },
        },
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole",
    )
    client.create_model_explainability_job_definition(
        JobDefinitionName="testmodelexplainabilityjobdefinition3",
        ModelExplainabilityAppSpecification={
            "ImageUri": "testimageuri",
            "ConfigUri": "testconfiguri",
        },
        ModelExplainabilityJobInput={
            "EndpointInput": {
                "EndpointName": "test3endpointname3",
                "LocalPath": "testlocalpath",
            },
        },
        ModelExplainabilityJobOutputConfig={
            "MonitoringOutputs": [
                {
                    "S3Output": {
                        "S3Uri": "s3://output/uri",
                        "LocalPath": "testlocalpath",
                    }
                },
            ]
        },
        JobResources={
            "ClusterConfig": {
                "InstanceCount": 1,
                "InstanceType": "ml.m4.xlarge",
                "VolumeSizeInGB": 4,
            },
        },
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole",
    )
    resp = client.list_model_explainability_job_definitions(
        EndpointName="testendpointname",
        SortBy="Name",
        SortOrder="Ascending",
        NameContains="testmodelexplainabilityjobdefinition",
        CreationTimeBefore=str(datetime.datetime(2099, 1, 1)),
        CreationTimeAfter=str(datetime.datetime(2021, 1, 1)),
    )["JobDefinitionSummaries"]
    assert len(resp) == 1
    assert (
        resp[0]["MonitoringJobDefinitionName"] == "testmodelexplainabilityjobdefinition"
    )
    assert (
        resp[0]["MonitoringJobDefinitionArn"]
        == f"arn:aws:sagemaker:eu-west-1:{ACCOUNT_ID}:model-explainability-job-definition/testmodelexplainabilityjobdefinition"
    )
    assert isinstance(resp[0]["CreationTime"], datetime.datetime)
    assert resp[0]["EndpointName"] == "testendpointname"


@mock_aws
def test_delete_model_explainability_job_definition():
    client = boto3.client("sagemaker", region_name="eu-west-1")
    client.create_model_explainability_job_definition(
        JobDefinitionName="testmodelexplainabilityjobdefinition",
        ModelExplainabilityAppSpecification={
            "ImageUri": "testimageuri",
            "ConfigUri": "testconfiguri",
        },
        ModelExplainabilityJobInput={
            "EndpointInput": {
                "EndpointName": "testendpointname",
                "LocalPath": "testlocalpath",
            },
        },
        ModelExplainabilityJobOutputConfig={
            "MonitoringOutputs": [
                {
                    "S3Output": {
                        "S3Uri": "s3://output/uri",
                        "LocalPath": "testlocalpath",
                    }
                },
            ]
        },
        JobResources={
            "ClusterConfig": {
                "InstanceCount": 1,
                "InstanceType": "ml.m4.xlarge",
                "VolumeSizeInGB": 4,
            },
        },
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole",
    )
    client.create_model_explainability_job_definition(
        JobDefinitionName="test2modelexplainabilityjobdefinition2",
        ModelExplainabilityAppSpecification={
            "ImageUri": "testimageuri",
            "ConfigUri": "testconfiguri",
        },
        ModelExplainabilityJobInput={
            "EndpointInput": {
                "EndpointName": "testendpointname",
                "LocalPath": "testlocalpath",
            },
        },
        ModelExplainabilityJobOutputConfig={
            "MonitoringOutputs": [
                {
                    "S3Output": {
                        "S3Uri": "s3://output/uri",
                        "LocalPath": "testlocalpath",
                    }
                },
            ]
        },
        JobResources={
            "ClusterConfig": {
                "InstanceCount": 1,
                "InstanceType": "ml.m4.xlarge",
                "VolumeSizeInGB": 4,
            },
        },
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole",
    )
    client.delete_model_explainability_job_definition(
        JobDefinitionName="testmodelexplainabilityjobdefinition"
    )
    resp = client.list_model_explainability_job_definitions()
    assert len(resp["JobDefinitionSummaries"]) == 1
    assert (
        resp["JobDefinitionSummaries"][0]["MonitoringJobDefinitionName"]
        == "test2modelexplainabilityjobdefinition2"
    )


@mock_aws
def test_tag_model_explainability_job_definition():
    client = boto3.client("sagemaker", region_name="eu-west-1")
    client.create_model_explainability_job_definition(
        JobDefinitionName="testmodelexplainabilityjobdefinition",
        ModelExplainabilityAppSpecification={
            "ImageUri": "testimageuri",
            "ConfigUri": "testconfiguri",
        },
        ModelExplainabilityJobInput={
            "EndpointInput": {
                "EndpointName": "testendpointname",
                "LocalPath": "testlocalpath",
            },
        },
        ModelExplainabilityJobOutputConfig={
            "MonitoringOutputs": [
                {
                    "S3Output": {
                        "S3Uri": "s3://output/uri",
                        "LocalPath": "testlocalpath",
                    }
                },
            ]
        },
        JobResources={
            "ClusterConfig": {
                "InstanceCount": 1,
                "InstanceType": "ml.m4.xlarge",
                "VolumeSizeInGB": 4,
            },
        },
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole",
        Tags=[{"Key": "testkey", "Value": "testvalue"}],
    )
    resp1 = client.list_tags(
        ResourceArn=f"arn:aws:sagemaker:eu-west-1:{ACCOUNT_ID}:model-explainability-job-definition/testmodelexplainabilityjobdefinition"
    )
    assert resp1["Tags"] == [{"Key": "testkey", "Value": "testvalue"}]
    client.add_tags(
        ResourceArn=f"arn:aws:sagemaker:eu-west-1:{ACCOUNT_ID}:model-explainability-job-definition/testmodelexplainabilityjobdefinition",
        Tags=[{"Key": "testkey2", "Value": "testvalue2"}],
    )
    resp2 = client.list_tags(
        ResourceArn=f"arn:aws:sagemaker:eu-west-1:{ACCOUNT_ID}:model-explainability-job-definition/testmodelexplainabilityjobdefinition"
    )
    assert resp2["Tags"] == [
        {"Key": "testkey", "Value": "testvalue"},
        {"Key": "testkey2", "Value": "testvalue2"},
    ]
    client.delete_tags(
        ResourceArn=f"arn:aws:sagemaker:eu-west-1:{ACCOUNT_ID}:model-explainability-job-definition/testmodelexplainabilityjobdefinition",
        TagKeys=["testkey"],
    )
    resp3 = client.list_tags(
        ResourceArn=f"arn:aws:sagemaker:eu-west-1:{ACCOUNT_ID}:model-explainability-job-definition/testmodelexplainabilityjobdefinition"
    )
    assert resp3["Tags"] == [{"Key": "testkey2", "Value": "testvalue2"}]



# mypy: ignore-errors
from collections import OrderedDict
from collections.abc import Mapping, MutableMapping

from botocore import xform_name
from botocore.utils import parse_timestamp

UNDEFINED = object()  # Sentinel to signal the absence of a field in the input


class QueryParser:
    TIMESTAMP_FORMAT = "iso8601"

    MAP_TYPE = dict

    def __init__(self, timestamp_parser=None, map_type=None):
        if timestamp_parser is None:
            timestamp_parser = parse_timestamp
        self._timestamp_parser = timestamp_parser
        if map_type is not None:
            self.MAP_TYPE = map_type

    def parse(self, request_dict, operation_model):
        shape = operation_model.input_shape
        parsed = self._do_parse(request_dict, shape)
        return parsed if parsed is not UNDEFINED else {}

    def _do_parse(self, request_dict, shape):
        parsed = self._parse_shape(shape, request_dict["query_params"])
        return parsed if parsed is not UNDEFINED else {}

    def _parse_shape(self, shape, node, prefix=""):
        handler = getattr(self, "_handle_%s" % shape.type_name, self._default_handle)
        return handler(shape, node, prefix)

    def _gonna_recurse(self, query_params, prefix):
        if prefix == "":
            return False
        return not any([param_key.startswith(prefix) for param_key in query_params])

    def _handle_structure(self, shape, query_params, prefix=""):
        if self._gonna_recurse(query_params, prefix):
            return UNDEFINED
        parsed = self.MAP_TYPE()
        members = shape.members
        for member_name in members:
            member_shape = members[member_name]
            member_prefix = self._get_serialized_name(member_shape, member_name)
            if prefix:
                member_prefix = "%s.%s" % (prefix, member_prefix)
            value = self._parse_shape(member_shape, query_params, member_prefix)
            parsed_key = self._parsed_key_name(member_name)
            if value is not UNDEFINED:
                parsed[parsed_key] = value
        return parsed if parsed != {} else UNDEFINED

    def _handle_list(self, shape, node, prefix=""):
        # The query protocol serializes empty lists as an empty string.
        value = self._parse_shape(shape.member, node, prefix)
        if value == "":
            return []

        list_name = shape.member.serialization.get("name", "member")
        list_prefix = f"{prefix}.{list_name}"
        parsed_list = []
        i = 1
        while True:
            element_name = f"{list_prefix}.{i}"
            element_shape = shape.member
            value = self._parse_shape(element_shape, node, element_name)
            if value is UNDEFINED:
                break
            parsed_list.append(value)
            i += 1
        return parsed_list if parsed_list != [] else UNDEFINED

    def _handle_timestamp(self, shape, query_params, prefix=""):
        value = self._default_handle(shape, query_params, prefix)
        return value if value is UNDEFINED else self._timestamp_parser(value)

    def _handle_boolean(self, shape, query_params, prefix=""):
        value = self._default_handle(shape, query_params, prefix)
        try:
            return value.lower() == "true"
        except AttributeError:
            pass
        return UNDEFINED

    def _handle_integer(self, shape, query_params, prefix=""):
        value = self._default_handle(shape, query_params, prefix)
        return value if value is UNDEFINED else int(value)

    def _handle_float(self, shape, query_params, prefix=""):
        value = self._default_handle(shape, query_params, prefix)
        return value if value is UNDEFINED else float(value)

    _handle_double = _handle_float
    _handle_long = _handle_integer

    def _default_handle(self, shape, value, prefix=""):
        default_value = shape.metadata.get("default", UNDEFINED)
        return value.get(prefix, default_value)

    def _get_serialized_name(self, shape, default_name):
        return shape.serialization.get("name", default_name)

    def _parsed_key_name(self, member_name):
        key_name = member_name
        return key_name


class XFormedDict(MutableMapping):
    """
    A Pascal/Snake case-insensitive  ``dict``-like object.

        xfd = XFormedDict()
        xfd['DBInstanceIdentifier'] = 'identifier'
        xfd['DBInstanceIdentifier'] == 'identifier'  # True
        xfd['db_instance_identifier'] == 'identifier'  # True
        list(xfd) == ['db_instance_identifier']  # True

    """

    def __init__(self, data=None, **kwargs):
        self._xform_cache = {}
        self._store = OrderedDict()
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def _xformed(self, key: str):
        return xform_name(key, _xform_cache=self._xform_cache)

    def __setitem__(self, key, value):
        # Use the xformed key for lookups, but store the actual
        # key alongside the value.
        self._store[self._xformed(key)] = (key, value)

    def __getitem__(self, key: str):
        return self._store[self._xformed(key)][1]

    def __delitem__(self, key):
        del self._store[self._xformed(key)]

    def __iter__(self):
        return (key for key in self._store.keys())

    def __len__(self):
        return len(self._store)

    def original_items(self):
        """Like iteritems(), but with all PascalCase keys."""
        return ((keyval[0], keyval[1]) for (_, keyval) in self._store.items())

    def __eq__(self, other):
        if isinstance(other, Mapping):
            other = XFormedDict(other)
        else:
            return NotImplemented
        # Compare xformed
        return dict(self.items()) == dict(other.items())

    def copy(self):
        return XFormedDict(self._store.values())

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, dict(self.items()))



"""Test different server responses."""

import moto.server as server


def test_emrserverless_list():
    backend = server.create_backend_app("emr-serverless")
    test_client = backend.test_client()

    resp = test_client.get("/applications")
    assert resp.status_code == 200
    assert "applications" in str(resp.data)



from typing import Any, Dict, List

from moto.stepfunctions.parser.asl.component.common.assign.assign_decl_binding import (
    AssignDeclBinding,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment


class AssignDecl(EvalComponent):
    declaration_bindings: List[AssignDeclBinding]

    def __init__(self, declaration_bindings: List[AssignDeclBinding]):
        super().__init__()
        self.declaration_bindings = declaration_bindings

    def _eval_body(self, env: Environment) -> None:
        declarations: Dict[str, Any] = dict()
        for declaration_binding in self.declaration_bindings:
            declaration_binding.eval(env=env)
            binding: Dict[str, Any] = env.stack.pop()
            declarations.update(binding)
        for identifier, value in declarations.items():
            env.variable_store.set(variable_identifier=identifier, variable_value=value)



from unittest.mock import patch

import pytest

from moto.s3.utils import (
    _VersionedKeyStore,
    bucket_name_from_url,
    compute_checksum,
    cors_matches_origin,
    parse_region_from_url,
)


def test_base_url():
    assert bucket_name_from_url("https://s3.amazonaws.com/") is None


def test_localhost_bucket():
    assert bucket_name_from_url("https://wfoobar.localhost:5000/abc") == "wfoobar"


def test_localhost_without_bucket():
    assert bucket_name_from_url("https://www.localhost:5000/def") is None


def test_force_ignore_subdomain_for_bucketnames():
    with patch("moto.s3.utils.S3_IGNORE_SUBDOMAIN_BUCKETNAME", True):
        assert (
            bucket_name_from_url("https://subdomain.localhost:5000/abc/resource")
            is None
        )


def test_versioned_key_store():
    key_store = _VersionedKeyStore()

    assert not key_store

    key_store["key"] = [1]
    assert len(key_store) == 1

    key_store["key"] = 2
    assert len(key_store) == 1

    assert key_store["key"] == 2
    assert key_store.get("key") == 2
    assert key_store.get("badkey") is None
    assert key_store.get("badkey", "HELLO") == "HELLO"

    # Tests key[
    assert "badkey" not in key_store
    with pytest.raises(KeyError):
        _ = key_store["badkey"]

    assert len(key_store.getlist("key")) == 2
    assert key_store.getlist("key") == [[1], 2]
    assert key_store.getlist("badkey") is None

    key_store.setlist("key", 1)
    assert key_store.getlist("key") == [1]

    key_store.setlist("key", (1, 2))
    assert key_store.getlist("key") != (1, 2)
    assert key_store.getlist("key") == [1, 2]

    key_store.setlist("key", [[1], [2]])
    assert len(key_store["key"]) == 1
    assert key_store.getlist("key") == [[1], [2]]


def test_parse_region_from_url():
    expected = "us-west-2"
    for url in [
        "http://s3-us-west-2.amazonaws.com/bucket",
        "http://s3.us-west-2.amazonaws.com/bucket",
        "http://bucket.s3-us-west-2.amazonaws.com",
        "https://s3-us-west-2.amazonaws.com/bucket",
        "https://s3.us-west-2.amazonaws.com/bucket",
        "https://bucket.s3-us-west-2.amazonaws.com",
    ]:
        assert parse_region_from_url(url) == expected

    expected = "us-east-1"
    for url in [
        "http://s3.amazonaws.com/bucket",
        "http://bucket.s3.amazonaws.com",
        "https://s3.amazonaws.com/bucket",
        "https://bucket.s3.amazonaws.com",
    ]:
        assert parse_region_from_url(url) == expected


def test_checksum_sha256():
    checksum = b"h9FJy0JMA4dlbyEdJYn7Wx4WIpkhMJ6YWIQZzMqKc2I="
    assert compute_checksum(b"somedata", "SHA256") == checksum
    # Unknown algorithms fallback to SHA256 for now
    assert compute_checksum(b"somedata", algorithm="unknown") == checksum


def test_checksum_sha1():
    assert compute_checksum(b"somedata", "SHA1") == b"76oxGuRIpzdMEiBhv+2VLZQOnjc="


def test_checksum_crc32():
    assert compute_checksum(b"somedata", "CRC32") == b"Uwy90A=="


def test_checksum_crc32c():
    try:
        import crc32c  # noqa # pylint: disable=unused-import

        assert compute_checksum(b"somedata", "CRC32C") == b"dB9qBQ=="
    except:  # noqa: E722 Do not use bare except
        # Optional library Can't be found - just revert to CRC32
        assert compute_checksum(b"somedata", "CRC32C") == b"Uwy90A=="


def test_cors_utils():
    "Fancy string matching"
    assert cors_matches_origin("a", ["a"])
    assert cors_matches_origin("b", ["a", "b"])
    assert not cors_matches_origin("c", [])
    assert not cors_matches_origin("c", ["a", "b"])

    assert cors_matches_origin("http://www.google.com", ["http://*.google.com"])
    assert cors_matches_origin("http://www.google.com", ["http://www.*.com"])
    assert cors_matches_origin("http://www.google.com", ["http://*"])
    assert cors_matches_origin("http://www.google.com", ["*"])

    assert not cors_matches_origin("http://www.google.com", ["http://www.*.org"])
    assert not cors_matches_origin("http://www.google.com", ["https://*"])



from typing import Any, Dict, Optional, Set

from botocore.exceptions import ClientError

from moto.stepfunctions.parser.api import HistoryEventType, TaskFailedEventDetails
from moto.stepfunctions.parser.asl.component.common.error_name.custom_error_name import (
    CustomErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.credentials import (
    ComputedCredentials,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.resource import (
    ResourceCondition,
    ResourceRuntimePart,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service_callback import (
    StateTaskServiceCallback,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.utils.boto_client import boto_client_for
from moto.stepfunctions.parser.asl.utils.encoding import to_json_str

_SUPPORTED_INTEGRATION_PATTERNS: Set[ResourceCondition] = {
    ResourceCondition.WaitForTaskToken,
}
_ERROR_NAME_CLIENT: str = "SQS.SdkClientException"
_ERROR_NAME_AWS: str = "SQS.AmazonSQSException"
_SUPPORTED_API_PARAM_BINDINGS: Dict[str, Set[str]] = {
    "sendmessage": {
        "DelaySeconds",
        "MessageAttributes",
        "MessageBody",
        "MessageDeduplicationId",
        "MessageGroupId",
        "QueueUrl",
    }
}


class StateTaskServiceSqs(StateTaskServiceCallback):
    def __init__(self):
        super().__init__(supported_integration_patterns=_SUPPORTED_INTEGRATION_PATTERNS)

    def _get_supported_parameters(self) -> Optional[Set[str]]:
        return _SUPPORTED_API_PARAM_BINDINGS.get(self.resource.api_action.lower())

    def _from_error(self, env: Environment, ex: Exception) -> FailureEvent:
        if isinstance(ex, ClientError):
            return FailureEvent(
                env=env,
                error_name=CustomErrorName(_ERROR_NAME_CLIENT),
                event_type=HistoryEventType.TaskFailed,
                event_details=EventDetails(
                    taskFailedEventDetails=TaskFailedEventDetails(
                        error=_ERROR_NAME_CLIENT,
                        cause=ex.response["Error"][
                            "Message"
                        ],  # TODO: update to report expected cause.
                        resource=self._get_sfn_resource(),
                        resourceType=self._get_sfn_resource_type(),
                    )
                ),
            )
        return super()._from_error(env=env, ex=ex)

    def _normalise_response(
        self,
        response: Any,
        boto_service_name: Optional[str] = None,
        service_action_name: Optional[str] = None,
    ) -> None:
        super()._normalise_response(
            response=response,
            boto_service_name=boto_service_name,
            service_action_name=service_action_name,
        )
        # Normalise output value keys to SFN standard for Md5OfMessageBody and Md5OfMessageAttributes
        if response and "Md5OfMessageBody" in response:
            md5_message_body = response.pop("Md5OfMessageBody")
            response["MD5OfMessageBody"] = md5_message_body

        if response and "Md5OfMessageAttributes" in response:
            md5_message_attributes = response.pop("Md5OfMessageAttributes")
            response["MD5OfMessageAttributes"] = md5_message_attributes

    def _eval_service_task(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
        task_credentials: ComputedCredentials,
    ):
        # TODO: Stepfunctions automatically dumps to json MessageBody's definitions.
        #  Are these other similar scenarios?
        if "MessageBody" in normalised_parameters:
            message_body = normalised_parameters["MessageBody"]
            if message_body is not None and not isinstance(message_body, str):
                normalised_parameters["MessageBody"] = to_json_str(message_body)

        service_name = self._get_boto_service_name()
        api_action = self._get_boto_service_action()
        sqs_client = boto_client_for(
            region=resource_runtime_part.region,
            account=resource_runtime_part.account,
            service=service_name,
            credentials=task_credentials,
        )
        response = getattr(sqs_client, api_action)(**normalised_parameters)
        response.pop("ResponseMetadata", None)
        env.stack.append(response)



from dataclasses import asdict, dataclass, field
from datetime import datetime
from typing import Any, Dict, Literal, Optional
from uuid import uuid4

from moto.appmesh.utils.common import clean_dict

Status = Dict[Literal["status"], str]


@dataclass
class Metadata:
    arn: str
    mesh_owner: str
    resource_owner: str
    created_at: datetime = datetime.now()
    last_updated_at: datetime = datetime.now()
    uid: str = uuid4().hex
    version: int = 1

    def update_timestamp(self) -> None:
        self.last_updated_at = datetime.now()


@dataclass
class Duration:
    unit: str
    value: int
    to_dict = asdict


class MissingField:
    def to_dict(self) -> None:
        return


@dataclass
class Timeout:
    idle: Optional[Duration] = field(default=None)
    per_request: Optional[Duration] = field(default=None)

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "idle": (self.idle or MissingField()).to_dict(),
                "perRequest": (self.per_request or MissingField()).to_dict(),
            }
        )



from moto.core.responses import BaseResponse

from .models import CognitoIdentityBackend, cognitoidentity_backends
from .utils import get_random_identity_id


class CognitoIdentityResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="cognito-identity")

    @property
    def backend(self) -> CognitoIdentityBackend:
        return cognitoidentity_backends[self.current_account][self.region]

    def create_identity_pool(self) -> str:
        identity_pool_name = self._get_param("IdentityPoolName")
        allow_unauthenticated_identities = self._get_param(
            "AllowUnauthenticatedIdentities"
        )
        supported_login_providers = self._get_param("SupportedLoginProviders")
        developer_provider_name = self._get_param("DeveloperProviderName")
        open_id_connect_provider_arns = self._get_param("OpenIdConnectProviderARNs")
        cognito_identity_providers = self._get_param("CognitoIdentityProviders")
        saml_provider_arns = self._get_param("SamlProviderARNs")
        pool_tags = self._get_param("IdentityPoolTags")

        return self.backend.create_identity_pool(
            identity_pool_name=identity_pool_name,
            allow_unauthenticated_identities=allow_unauthenticated_identities,
            supported_login_providers=supported_login_providers,
            developer_provider_name=developer_provider_name,
            open_id_connect_provider_arns=open_id_connect_provider_arns,
            cognito_identity_providers=cognito_identity_providers,
            saml_provider_arns=saml_provider_arns,
            tags=pool_tags,
        )

    def update_identity_pool(self) -> str:
        pool_id = self._get_param("IdentityPoolId")
        pool_name = self._get_param("IdentityPoolName")
        allow_unauthenticated = self._get_bool_param("AllowUnauthenticatedIdentities")
        login_providers = self._get_param("SupportedLoginProviders")
        provider_name = self._get_param("DeveloperProviderName")
        provider_arns = self._get_param("OpenIdConnectProviderARNs")
        identity_providers = self._get_param("CognitoIdentityProviders")
        saml_providers = self._get_param("SamlProviderARNs")
        pool_tags = self._get_param("IdentityPoolTags")

        return self.backend.update_identity_pool(
            identity_pool_id=pool_id,
            identity_pool_name=pool_name,
            allow_unauthenticated=allow_unauthenticated,
            login_providers=login_providers,
            provider_name=provider_name,
            provider_arns=provider_arns,
            identity_providers=identity_providers,
            saml_providers=saml_providers,
            tags=pool_tags,
        )

    def get_id(self) -> str:
        return self.backend.get_id(identity_pool_id=self._get_param("IdentityPoolId"))

    def describe_identity_pool(self) -> str:
        return self.backend.describe_identity_pool(self._get_param("IdentityPoolId"))

    def get_credentials_for_identity(self) -> str:
        return self.backend.get_credentials_for_identity(self._get_param("IdentityId"))

    def get_open_id_token_for_developer_identity(self) -> str:
        return self.backend.get_open_id_token_for_developer_identity(
            self._get_param("IdentityId") or get_random_identity_id(self.region)
        )

    def get_open_id_token(self) -> str:
        return self.backend.get_open_id_token(
            self._get_param("IdentityId") or get_random_identity_id(self.region)
        )

    def list_identities(self) -> str:
        return self.backend.list_identities(
            self._get_param("IdentityPoolId") or get_random_identity_id(self.region)
        )

    def list_identity_pools(self) -> str:
        return self.backend.list_identity_pools()

    def delete_identity_pool(self) -> str:
        identity_pool_id = self._get_param("IdentityPoolId")
        self.backend.delete_identity_pool(
            identity_pool_id=identity_pool_id,
        )
        return ""



import json
from unittest import SkipTest

import boto3
import requests

from moto import mock_aws, settings

from ..test_awslambda.utilities import get_role_name, get_test_zip_file1

LAMBDA_REGION = "us-west-2"
PYTHON_VERSION = "3.11"
FUNCTION_NAME = "test-function-123"

if settings.TEST_SERVER_MODE:
    raise SkipTest("No point in testing batch_simple in ServerMode")


@mock_aws(config={"lambda": {"use_docker": False}})
def test_run_function():
    # Setup
    client = setup_lambda()

    # Execute
    result = client.invoke(
        FunctionName=FUNCTION_NAME,
        LogType="Tail",
    )

    # Verify
    assert result["StatusCode"] == 200
    assert result["Payload"].read().decode("utf-8") == "Simple Lambda happy path OK"


@mock_aws(config={"lambda": {"use_docker": False}})
def test_run_function_no_log():
    # Setup
    client = setup_lambda()
    payload = {"results": "results"}

    # Execute
    result = client.invoke(FunctionName=FUNCTION_NAME, Payload=json.dumps(payload))

    # Verify
    assert result["StatusCode"] == 200
    assert json.loads(result["Payload"].read().decode("utf-8")) == payload

    # Execute
    result = client.invoke(FunctionName=FUNCTION_NAME)

    # Verify
    assert result["StatusCode"] == 200
    assert result["Payload"].read().decode("utf-8") == "Simple Lambda happy path OK"


@mock_aws(config={"lambda": {"use_docker": False}})
def test_set_lambda_simple_query_results():
    # Setup
    base_url = (
        settings.test_server_mode_endpoint()
        if settings.TEST_SERVER_MODE
        else "http://motoapi.amazonaws.com"
    )
    results = {"results": ["test", "test 2"], "region": LAMBDA_REGION}
    resp = requests.post(
        f"{base_url}/moto-api/static/lambda-simple/response",
        json=results,
    )
    assert resp.status_code == 201

    client = setup_lambda()

    # Execute & Verify
    resp = client.invoke(
        FunctionName=FUNCTION_NAME,
        LogType="Tail",
    )
    assert resp["Payload"].read().decode() == results["results"][0]

    resp = client.invoke(
        FunctionName=FUNCTION_NAME,
        LogType="Tail",
    )
    assert resp["Payload"].read().decode() == results["results"][1]

    resp = client.invoke(
        FunctionName=FUNCTION_NAME,
        LogType="Tail",
    )
    assert resp["Payload"].read().decode() == "Simple Lambda happy path OK"


def setup_lambda():
    client = boto3.client("lambda", LAMBDA_REGION)
    zip_content = get_test_zip_file1()
    function_name = FUNCTION_NAME
    role = get_role_name()
    client.create_function(
        FunctionName=function_name,
        Runtime=PYTHON_VERSION,
        Role=role,
        Handler="lambda_function.lambda_handler",
        Code={"ZipFile": zip_content},
        Description="test lambda function",
        Timeout=3,
        MemorySize=128,
        Publish=True,
    )
    return client



from typing import Any

from moto.stepfunctions.parser.api import ExecutionFailedEventDetails, HistoryEventType
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
    FailureEventException,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name import (
    StatesErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)
from moto.stepfunctions.parser.asl.component.common.variable_sample import (
    VariableSample,
)
from moto.stepfunctions.parser.asl.component.state.wait.wait_function.wait_function import (
    WaitFunction,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.utils.json_path import extract_json


class SecondsPath(WaitFunction):
    # SecondsPath
    # A time, in seconds, to state_wait before beginning the state specified in the Next
    # field, specified using a path from the state's input data.
    # You must specify an integer value for this field.

    def __init__(self, path: str):
        self.path: str = path

    def _validate_seconds_value(self, env: Environment, seconds: Any):
        if isinstance(seconds, int) and seconds >= 0:
            return
        error_type = StatesErrorNameType.StatesRuntime

        assignment_description = f"{self.path} == {seconds}"
        if not isinstance(seconds, int):
            cause = f"The SecondsPath parameter cannot be parsed as a long value: {assignment_description}"
        else:  # seconds < 0
            cause = f"The SecondsPath parameter references a negative value: {assignment_description}"

        raise FailureEventException(
            failure_event=FailureEvent(
                env=env,
                error_name=StatesErrorName(typ=error_type),
                event_type=HistoryEventType.ExecutionFailed,
                event_details=EventDetails(
                    executionFailedEventDetails=ExecutionFailedEventDetails(
                        error=error_type.to_name(), cause=cause
                    )
                ),
            )
        )

    def _get_wait_seconds(self, env: Environment) -> int:
        inp = env.stack[-1]
        seconds = extract_json(self.path, inp)
        self._validate_seconds_value(env=env, seconds=seconds)
        return seconds


class SecondsPathVar(SecondsPath):
    variable_sample: VariableSample

    def __init__(self, variable_sample: VariableSample):
        super().__init__(path=variable_sample.expression)
        self.variable_sample = variable_sample

    def _get_wait_seconds(self, env: Environment) -> int:
        self.variable_sample.eval(env=env)
        seconds = env.stack.pop()
        self._validate_seconds_value(env=env, seconds=seconds)
        return seconds



"""Directory-related unit tests common to different directory types.

Simple AD directories are used for test data, but the operations are
common to the other directory types.
"""

from datetime import datetime, timezone

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws
from moto.moto_api._internal import mock_random

from .test_ds_simple_ad_directory import TEST_REGION, create_test_directory


@mock_aws
def test_ds_delete_directory():
    """Test good and bad invocations of delete_directory()."""
    client = boto3.client("ds", region_name=TEST_REGION)

    # Delete a directory when there are none.
    random_directory_id = f"d-{mock_random.get_random_hex(10)}"
    with pytest.raises(ClientError) as exc:
        client.delete_directory(DirectoryId=random_directory_id)
    err = exc.value.response["Error"]
    assert err["Code"] == "EntityDoesNotExistException"
    assert f"Directory {random_directory_id} does not exist" in err["Message"]

    # Delete an existing directory.
    ec2_client = boto3.client("ec2", region_name=TEST_REGION)
    directory_id = create_test_directory(client, ec2_client)
    result = client.delete_directory(DirectoryId=directory_id)
    assert result["DirectoryId"] == directory_id

    # Verify there are no dictionaries, network interfaces or associated
    # security groups.
    result = client.describe_directories()
    assert len(result["DirectoryDescriptions"]) == 0
    result = ec2_client.describe_network_interfaces()
    assert len(result["NetworkInterfaces"]) == 0
    result = ec2_client.describe_security_groups()
    for group in result["SecurityGroups"]:
        assert "directory controllers" not in group["Description"]

    # Attempt to delete a non-existent directory.
    nonexistent_id = f"d-{mock_random.get_random_hex(10)}"
    with pytest.raises(ClientError) as exc:
        client.delete_directory(DirectoryId=nonexistent_id)
    err = exc.value.response["Error"]
    assert err["Code"] == "EntityDoesNotExistException"
    assert f"Directory {nonexistent_id} does not exist" in err["Message"]

    # Attempt to use an invalid directory ID.
    bad_id = mock_random.get_random_hex(3)
    with pytest.raises(ClientError) as exc:
        client.delete_directory(DirectoryId=bad_id)
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert "1 validation error detected" in err["Message"]
    assert (
        f"Value '{bad_id}' at 'directoryId' failed to satisfy constraint: "
        f"Member must satisfy regular expression pattern: ^d-[0-9a-f]{{10}}$"
    ) in err["Message"]


@mock_aws
def test_ds_get_directory_limits():
    """Test return value for directory limits."""
    client = boto3.client("ds", region_name=TEST_REGION)
    ec2_client = boto3.client("ec2", region_name=TEST_REGION)

    limits = client.get_directory_limits()["DirectoryLimits"]
    assert limits["CloudOnlyDirectoriesCurrentCount"] == 0
    assert limits["CloudOnlyDirectoriesLimit"] > 0
    assert not limits["CloudOnlyDirectoriesLimitReached"]

    # Create a bunch of directories and verify the current count has been
    # updated.
    for _ in range(limits["CloudOnlyDirectoriesLimit"]):
        create_test_directory(client, ec2_client)
    limits = client.get_directory_limits()["DirectoryLimits"]
    assert (
        limits["CloudOnlyDirectoriesLimit"]
        == limits["CloudOnlyDirectoriesCurrentCount"]
    )
    assert limits["CloudOnlyDirectoriesLimitReached"]
    assert not limits["CloudOnlyMicrosoftADCurrentCount"]
    assert not limits["ConnectedDirectoriesCurrentCount"]


@mock_aws
def test_ds_describe_directories():
    """Test good and bad invocations of describe_directories()."""
    client = boto3.client("ds", region_name=TEST_REGION)
    ec2_client = boto3.client("ec2", region_name=TEST_REGION)

    expected_ids = set()
    limit = 10
    for _ in range(limit):
        expected_ids.add(create_test_directory(client, ec2_client))

    # Test that if no directory IDs are specified, all are returned.
    result = client.describe_directories()
    directories = result["DirectoryDescriptions"]
    directory_ids = [x["DirectoryId"] for x in directories]

    assert len(directories) == limit
    assert set(directory_ids) == expected_ids
    for idx, dir_info in enumerate(directories):
        assert dir_info["DesiredNumberOfDomainControllers"] == 0
        assert not dir_info["SsoEnabled"]
        assert dir_info["DirectoryId"] == directory_ids[idx]
        assert dir_info["Name"].startswith("test-")
        assert dir_info["Size"] == "Large"
        assert dir_info["Alias"] == directory_ids[idx]
        assert dir_info["AccessUrl"] == f"{directory_ids[idx]}.awsapps.com"
        assert dir_info["Stage"] == "Active"
        assert dir_info["LaunchTime"] <= datetime.now(timezone.utc)
        assert dir_info["StageLastUpdatedDateTime"] <= datetime.now(timezone.utc)
        assert dir_info["Type"] == "SimpleAD"
        assert dir_info["VpcSettings"]["VpcId"].startswith("vpc-")
        assert len(dir_info["VpcSettings"]["SubnetIds"]) == 2
        assert dir_info["VpcSettings"]["SecurityGroupId"].startswith("sg-")
        assert len(dir_info["DnsIpAddrs"]) == 2
    assert "NextToken" not in result

    # Test with a specific directory ID.
    result = client.describe_directories(DirectoryIds=[directory_ids[5]])
    assert len(result["DirectoryDescriptions"]) == 1
    assert result["DirectoryDescriptions"][0]["DirectoryId"] == directory_ids[5]

    # Test with a bad directory ID.
    bad_id = mock_random.get_random_hex(3)
    with pytest.raises(ClientError) as exc:
        client.describe_directories(DirectoryIds=[bad_id])
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        f"Value '{bad_id}' at 'directoryId' failed to satisfy constraint: "
        f"Member must satisfy regular expression pattern: ^d-[0-9a-f]{{10}}$"
    ) in err["Message"]

    # Test with an invalid next token.
    with pytest.raises(ClientError) as exc:
        client.describe_directories(NextToken="bogus")
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidNextTokenException"
    assert "Invalid value passed for the NextToken parameter" in err["Message"]

    # Test with a limit.
    result = client.describe_directories(Limit=5)
    assert len(result["DirectoryDescriptions"]) == 5
    directories = result["DirectoryDescriptions"]
    for idx in range(5):
        assert directories[idx]["DirectoryId"] == directory_ids[idx]
    assert result["NextToken"]

    result = client.describe_directories(Limit=1, NextToken=result["NextToken"])
    assert len(result["DirectoryDescriptions"]) == 1
    assert result["DirectoryDescriptions"][0]["DirectoryId"] == directory_ids[5]


@mock_aws
def test_ds_create_alias():
    """Test good and bad invocations of create_alias()."""
    client = boto3.client("ds", region_name=TEST_REGION)
    ec2_client = boto3.client("ec2", region_name=TEST_REGION)

    # Create a directory we can test against.
    directory_id = create_test_directory(client, ec2_client)

    # Bad format.
    bad_alias = f"d-{mock_random.get_random_hex(10)}"
    with pytest.raises(ClientError) as exc:
        client.create_alias(DirectoryId=directory_id, Alias=bad_alias)
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        rf"Value '{bad_alias}' at 'alias' failed to satisfy constraint: "
        rf"Member must satisfy regular expression pattern: "
        rf"^(?!D-|d-)([\da-zA-Z]+)([-]*[\da-zA-Z])*$"
    ) in err["Message"]

    # Too long.
    bad_alias = f"d-{mock_random.get_random_hex(62)}"
    with pytest.raises(ClientError) as exc:
        client.create_alias(DirectoryId=directory_id, Alias=bad_alias)
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        f"Value '{bad_alias}' at 'alias' failed to satisfy constraint: "
        f"Member must have length less than or equal to 62"
    ) in err["Message"]

    # Just right.
    good_alias = f"{mock_random.get_random_hex(10)}"
    result = client.create_alias(DirectoryId=directory_id, Alias=good_alias)
    assert result["DirectoryId"] == directory_id
    assert result["Alias"] == good_alias
    result = client.describe_directories()
    directory = result["DirectoryDescriptions"][0]
    assert directory["Alias"] == good_alias
    assert directory["AccessUrl"] == f"{good_alias}.awsapps.com"

    # Attempt to create another alias for the same directory.
    another_good_alias = f"{mock_random.get_random_hex(10)}"
    with pytest.raises(ClientError) as exc:
        client.create_alias(DirectoryId=directory_id, Alias=another_good_alias)
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidParameterException"
    assert (
        "The directory in the request already has an alias. That alias must "
        "be deleted before a new alias can be created."
    ) in err["Message"]

    # Create a second directory we can test against.
    directory_id2 = create_test_directory(client, ec2_client)
    with pytest.raises(ClientError) as exc:
        client.create_alias(DirectoryId=directory_id2, Alias=good_alias)
    err = exc.value.response["Error"]
    assert err["Code"] == "EntityAlreadyExistsException"
    assert f"Alias '{good_alias}' already exists." in err["Message"]


@mock_aws
def test_ds_enable_sso():
    """Test good and bad invocations of enable_sso()."""
    client = boto3.client("ds", region_name=TEST_REGION)
    ec2_client = boto3.client("ec2", region_name=TEST_REGION)

    # Create a directory we can test against.
    directory_id = create_test_directory(client, ec2_client)

    # Need an alias before setting SSO.
    with pytest.raises(ClientError) as exc:
        client.enable_sso(DirectoryId=directory_id)
    err = exc.value.response["Error"]
    assert err["Code"] == "ClientException"
    assert (
        f"An alias is required before enabling SSO. DomainId={directory_id}"
    ) in err["Message"]

    # Add the alias to continue testing.
    client.create_alias(DirectoryId=directory_id, Alias="anything-goes")

    # Password must be less than 128 chars in length.
    good_username = "test"
    bad_password = f"bad_password{mock_random.get_random_hex(128)}"
    with pytest.raises(ClientError) as exc:
        client.enable_sso(
            DirectoryId=directory_id, UserName=good_username, Password=bad_password
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        "Value at 'ssoPassword' failed to satisfy constraint: Member must "
        "have length less than or equal to 128"
    ) in err["Message"]

    # Username has constraints.
    bad_username = "@test"
    good_password = "password"
    with pytest.raises(ClientError) as exc:
        client.enable_sso(
            DirectoryId=directory_id, UserName=bad_username, Password=good_password
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        rf"Value '{bad_username}' at 'userName' failed to satisfy constraint: "
        rf"Member must satisfy regular expression pattern: ^[a-zA-Z0-9._-]+$"
    ) in err["Message"]

    # Valid execution.
    client.enable_sso(DirectoryId=directory_id)
    result = client.describe_directories()
    directory = result["DirectoryDescriptions"][0]
    assert directory["SsoEnabled"]


@mock_aws
def test_ds_disable_sso():
    """Test good and bad invocations of disable_sso()."""
    client = boto3.client("ds", region_name=TEST_REGION)
    ec2_client = boto3.client("ec2", region_name=TEST_REGION)

    # Create a directory we can test against.
    directory_id = create_test_directory(client, ec2_client)

    # Password must be less than 128 chars in length.
    good_username = "test"
    bad_password = f"bad_password{mock_random.get_random_hex(128)}"
    with pytest.raises(ClientError) as exc:
        client.disable_sso(
            DirectoryId=directory_id, UserName=good_username, Password=bad_password
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        "Value at 'ssoPassword' failed to satisfy constraint: Member must "
        "have length less than or equal to 128"
    ) in err["Message"]

    # Username has constraints.
    bad_username = "@test"
    good_password = "password"
    with pytest.raises(ClientError) as exc:
        client.disable_sso(
            DirectoryId=directory_id, UserName=bad_username, Password=good_password
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        rf"Value '{bad_username}' at 'userName' failed to satisfy constraint: "
        rf"Member must satisfy regular expression pattern: ^[a-zA-Z0-9._-]+$"
    ) in err["Message"]

    # Valid execution.  First enable SSO, as the default is disabled SSO.
    client.create_alias(DirectoryId=directory_id, Alias="anything-goes")
    client.enable_sso(DirectoryId=directory_id)
    client.disable_sso(DirectoryId=directory_id)
    result = client.describe_directories()
    directory = result["DirectoryDescriptions"][0]
    assert not directory["SsoEnabled"]


@mock_aws
def test_create_trust():
    client = boto3.client("ds", region_name=TEST_REGION)
    ec2_client = boto3.client("ec2", region_name=TEST_REGION)
    directory_id = create_test_directory(client, ec2_client)

    # Create a trust
    trust_id = client.create_trust(
        DirectoryId=directory_id,
        RemoteDomainName="example.com",
        TrustPassword="password",
        TrustDirection="One-Way: Outgoing",
        TrustType="External",
    )["TrustId"]

    assert trust_id is not None

    # Describe the trust
    trusts = client.describe_trusts(DirectoryId=directory_id)["Trusts"]
    assert len(trusts) == 1
    assert trusts[0]["TrustId"] == trust_id
    assert trusts[0]["DirectoryId"] == directory_id


@mock_aws
def test_describe_trusts():
    client = boto3.client("ds", region_name=TEST_REGION)
    ec2_client = boto3.client("ec2", region_name=TEST_REGION)
    directory_id = create_test_directory(client, ec2_client)

    # Create several trusts
    trust_ids = []
    for x in range(5):
        trust_id = client.create_trust(
            DirectoryId=directory_id,
            RemoteDomainName=f"example{x}.com",
            TrustPassword="P@ssword1234!",
            TrustDirection="One-Way: Outgoing",
            TrustType="External",
        )["TrustId"]
        trust_ids.append(trust_id)

    # Describe the trusts
    trusts = client.describe_trusts(DirectoryId=directory_id)["Trusts"]
    assert len(trusts) == 5
    for idx, trust in enumerate(trusts):
        assert trust["TrustId"] == trust_ids[idx]
        assert trust["DirectoryId"] == directory_id

    # Describe a single trust
    trust = client.describe_trusts(DirectoryId=directory_id, TrustIds=[trust_ids[2]])[
        "Trusts"
    ][0]
    assert trust["TrustId"] == trust_ids[2]

    # Describe multiple trusts
    trusts = client.describe_trusts(DirectoryId=directory_id, TrustIds=trust_ids[1:3])[
        "Trusts"
    ]
    assert len(trusts) == 2
    assert trusts[0]["TrustId"] == trust_ids[1]
    assert trusts[1]["TrustId"] == trust_ids[2]

    # Describe all the trusts in the account
    trusts = client.describe_trusts()["Trusts"]
    assert len(trusts) == 5


@mock_aws
def test_delete_trust():
    client = boto3.client("ds", region_name=TEST_REGION)
    ec2_client = boto3.client("ec2", region_name=TEST_REGION)
    directory_id = create_test_directory(client, ec2_client)

    # Create several trusts
    trust_ids = []
    for x in range(2):
        trust_id = client.create_trust(
            DirectoryId=directory_id,
            RemoteDomainName=f"example{x}.com",
            TrustPassword="P@ssword1234!",
            TrustDirection="One-Way: Outgoing",
            TrustType="External",
        )["TrustId"]
        trust_ids.append(trust_id)

    # Verify the expected trusts exist
    trusts = client.describe_trusts(DirectoryId=directory_id)["Trusts"]
    assert len(trusts) == 2

    # Delete a trust
    resp = client.delete_trust(TrustId=trust_ids[0])
    assert resp["TrustId"] == trust_ids[0]

    # Verify the trust was deleted
    trusts = client.describe_trusts(DirectoryId=directory_id)["Trusts"]
    assert len(trusts) == 1
    assert trusts[0]["TrustId"] != trust_ids[0]

    # Test deleting a trust that doesn't exist
    with pytest.raises(ClientError) as exc:
        client.delete_trust(TrustId="t-1234567890")
    err = exc.value.response["Error"]
    assert err["Code"] == "EntityDoesNotExistException"


@mock_aws
def test_ldaps_exceptions_non_microsoftad():
    """Test LDAPS operations on non-Microsoft AD directories."""

    client = boto3.client("ds", region_name=TEST_REGION)
    ec2_client = boto3.client("ec2", region_name=TEST_REGION)

    directory_id = create_test_directory(client, ec2_client)

    # Test enabling LDAPS on a non-Microsoft AD directory.
    with pytest.raises(ClientError) as exc:
        client.enable_ldaps(DirectoryId=directory_id, Type="Client")
    err = exc.value.response["Error"]
    assert err["Code"] == "UnsupportedOperationException"

    # Test describe_ldaps_settings on a non-Microsoft AD directory.
    with pytest.raises(ClientError) as exc:
        client.describe_ldaps_settings(DirectoryId=directory_id, Type="Client")
    err = exc.value.response["Error"]
    assert err["Code"] == "UnsupportedOperationException"

    # Test disable_ldaps on a non-Microsoft AD directory.
    with pytest.raises(ClientError) as exc:
        client.disable_ldaps(DirectoryId=directory_id, Type="Client")
    err = exc.value.response["Error"]
    assert err["Code"] == "UnsupportedOperationException"


@mock_aws
def test_settings_exception_non_microsoftad():
    """Test Settings operations on non-Microsoft AD directories."""

    client = boto3.client("ds", region_name=TEST_REGION)
    ec2_client = boto3.client("ec2", region_name=TEST_REGION)

    directory_id = create_test_directory(client, ec2_client)
    # Test describing Settings on a non-Microsoft AD directory.
    with pytest.raises(ClientError) as exc:
        client.describe_settings(DirectoryId=directory_id)
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidParameterException"



from __future__ import annotations

import copy
import datetime
import re
from collections import OrderedDict, namedtuple
from enum import Enum
from functools import lru_cache
from typing import Any, Dict, List, Optional, Tuple

from botocore.loaders import create_loader
from botocore.model import ServiceModel
from botocore.utils import merge_dicts

SECONDS_IN_ONE_DAY = 24 * 60 * 60
FilterDef = namedtuple(
    "FilterDef",
    [
        # A list of object attributes to check against the filter values.
        # Set to None if filter is not yet implemented in `moto`.
        "attrs_to_check",
        # Description of the filter, e.g. 'Object Identifiers'.
        # Used in filter error messaging.
        "description",
    ],
)


class DbInstanceEngine(str, Enum):
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/rds/client/create_db_instance.html
    # 2023-11-08
    AURORA_MYSQL = "aurora-mysql"
    AURORA_POSTGRESQL = "aurora-postgresql"
    CUSTOM_ORACLE_EE = "custom-oracle-ee"
    CUSTOM_ORACLE_EE_CDB = "custom-oracle-ee-cdb"
    CUSTOM_SQLSERVER_EE = "custom-sqlserver-ee"
    CUSTOM_SQLSERVER_SE = "custom-sqlserver-se"
    CUSTOM_SQLSERVER_WEB = "custom-sqlserver-web"
    MARIADB = "mariadb"
    MYSQL = "mysql"
    ORACLE_EE = "oracle-ee"
    ORACLE_EE_CDB = "oracle-ee-cdb"
    ORACLE_SE2 = "oracle-se2"
    ORACLE_SE2_CDB = "oracle-se2-cdb"
    POSTGRES = "postgres"
    SQLSERVER_EE = "sqlserver-ee"
    SQLSERVER_SE = "sqlserver-se"
    SQLSERVER_EX = "sqlserver-ex"
    SQLSERVER_WEB = "sqlserver-web"

    @classmethod
    def valid_db_instance_engine(self) -> List[str]:
        return sorted([item.value for item in DbInstanceEngine])


class ClusterEngine(str, Enum):
    AURORA_POSTGRESQL = "aurora-postgresql"
    AURORA_MYSQL = "aurora-mysql"
    NEPTUNE = "neptune"
    RDS_POSTGRESQL = "postgres"
    RDS_MYSQL = "mysql"

    @classmethod
    def list_cluster_engines(self) -> List[str]:
        return sorted([item.value for item in ClusterEngine])

    @classmethod
    def serverless_engines(self) -> List[str]:
        return [ClusterEngine.AURORA_MYSQL, ClusterEngine.AURORA_POSTGRESQL]


def get_object_value(obj: Any, attr: str) -> Any:
    """Retrieves an arbitrary attribute value from an object.

    Nested attributes can be specified using dot notation,
    e.g. 'parent.child'.

    :param object obj:
        A valid Python object.
    :param str attr:
        The attribute name of the value to retrieve from the object.
    :returns:
        The attribute value, if it exists, or None.
    :rtype:
        any
    """
    keys = attr.split(".")
    val = obj
    for key in keys:
        if hasattr(val, key):
            val = getattr(val, key)
        else:
            return None
    return val


def merge_filters(
    filters_to_update: Optional[Dict[str, Any]], filters_to_merge: Dict[str, Any]
) -> Dict[str, Any]:
    """Given two groups of filters, merge the second into the first.

    List values are appended instead of overwritten:

    >>> merge_filters({'filter-name': ['value1']}, {'filter-name':['value2']})
    >>> {'filter-name': ['value1', 'value2']}

    :param filters_to_update:
        The filters to update.
    :type filters_to_update:
        dict[str, list] or None
    :param filters_to_merge:
        The filters to merge.
    :type filters_to_merge:
        dict[str, list] or None
    :returns:
        The updated filters.
    :rtype:
        dict[str, list]
    """
    if filters_to_update is None:
        filters_to_update = {}
    if filters_to_merge is None:
        filters_to_merge = {}
    merge_dicts(filters_to_update, filters_to_merge, append_lists=True)
    return filters_to_update


def validate_filters(
    filters: Dict[str, Any], filter_defs: Dict[str, FilterDef]
) -> None:
    """Validates filters against a set of filter definitions.

    Raises standard Python exceptions which should be caught
    and translated to an appropriate AWS/Moto exception higher
    up the call stack.

    :param dict[str, list] filters:
        The filters to validate.
    :param dict[str, FilterDef] filter_defs:
        The filter definitions to validate against.
    :returns: None
    :rtype: None
    :raises KeyError:
        if filter name not found in the filter definitions.
    :raises ValueError:
        if filter values is an empty list.
    :raises NotImplementedError:
        if `moto` does not yet support this filter.
    """
    for filter_name, filter_values in filters.items():
        filter_def = filter_defs.get(filter_name)
        if filter_def is None:
            raise KeyError(f"Unrecognized filter name: {filter_name}")
        if not filter_values:
            raise ValueError(f"The list of {filter_def.description} must not be empty.")
        if filter_def.attrs_to_check is None:
            raise NotImplementedError(
                f"{filter_name} filter has not been implemented in Moto yet."
            )


def apply_filter(resources: Any, filters: Any, filter_defs: Any) -> Any:
    """Apply an arbitrary filter to a group of resources.

    :param dict[str, object] resources:
        A dictionary mapping resource identifiers to resource objects.
    :param dict[str, list] filters:
        The filters to apply.
    :param dict[str, FilterDef] filter_defs:
        The supported filter definitions for the resource type.
    :returns:
        The filtered collection of resources.
    :rtype:
        dict[str, object]
    """
    resources_filtered = OrderedDict()
    for identifier, obj in resources.items():
        matches_filter = False
        for filter_name, filter_values in filters.items():
            filter_def = filter_defs.get(filter_name)
            for attr in filter_def.attrs_to_check:
                if get_object_value(obj, attr) in filter_values:
                    matches_filter = True
                    break
            else:
                matches_filter = False
            if not matches_filter:
                break
        if matches_filter:
            resources_filtered[identifier] = obj
    return resources_filtered


def get_start_date_end_date(
    base_date: str, window: str
) -> Tuple[datetime.datetime, datetime.datetime]:
    """Gets the start date and end date given DDD:HH24:MM-DDD:HH24:MM.

    :param base_date:
        type datetime
    :param window:
        DDD:HH24:MM-DDD:HH24:MM
    :returns:
        Start and End Date in datetime format
    :rtype:
        tuple
    """
    days = {"mon": 1, "tue": 2, "wed": 3, "thu": 4, "fri": 5, "sat": 6, "sun": 7}
    start = datetime.datetime.strptime(
        base_date + " " + window[4:9], "%d/%m/%y %H:%M"
    ) + datetime.timedelta(days=days[window[0:3]])
    end = datetime.datetime.strptime(
        base_date + " " + window[14::], "%d/%m/%y %H:%M"
    ) + datetime.timedelta(days=days[window[10:13]])
    return start, end


def get_start_date_end_date_from_time(
    base_date: str, window: str
) -> Tuple[datetime.datetime, datetime.datetime, bool]:
    """Gets the start date and end date given HH24:MM-HH24:MM.

    :param window:
        HH24:MM-HH24:MM
    :returns:
        Start and End Date in datetime format
        along with flag for spills over a day
        This is useful when determine time overlaps
    :rtype:
        tuple
    """
    times = window.split("-")
    spillover = False
    start = datetime.datetime.strptime(base_date + " " + times[0], "%d/%m/%y %H:%M")
    end = datetime.datetime.strptime(base_date + " " + times[1], "%d/%m/%y %H:%M")
    if end < start:
        end += datetime.timedelta(days=1)
        spillover = True
    return start, end, spillover


def get_overlap_between_two_date_ranges(
    start_time_1: datetime.datetime,
    end_time_1: datetime.datetime,
    start_time_2: datetime.datetime,
    end_time_2: datetime.datetime,
) -> int:
    """
    Determines overlap between 2 date ranges. Returns the overlap in seconds.
    """
    latest_start = max(start_time_1, start_time_2)
    earliest_end = min(end_time_1, end_time_2)
    delta = earliest_end - latest_start
    return (delta.days * SECONDS_IN_ONE_DAY) + delta.seconds


def valid_preferred_maintenance_window(
    maintenance_window: Any, backup_window: Any
) -> Optional[str]:
    """Determines validity of preferred_maintenance_window

    :param maintenance_windown:
        type DDD:HH24:MM-DDD:HH24:MM
    :param backup_window:
        type HH24:MM-HH24:MM
    :returns:
        message
    :rtype:
        str
    """
    MINUTES_30 = 1800
    HOURS_24 = 86400
    base_date = datetime.datetime.now().strftime("%d/%m/%y")
    try:
        p = re.compile(
            "([a-z]{3}):([0-9]{2}):([0-9]{2})-([a-z]{3}):([0-9]{2}):([0-9]{2})"
        )
        if len(maintenance_window) != 19 or re.search(p, maintenance_window) is None:
            return f"Invalid maintenance window format: {maintenance_window}. Should be specified as a range ddd:hh24:mi-ddd:hh24:mi (24H Clock UTC). Example: Sun:23:45-Mon:00:15"
        if backup_window:
            (
                backup_window_start,
                backup_window_end,
                backup_spill,
            ) = get_start_date_end_date_from_time(base_date, backup_window)
            (
                maintenance_window_start,
                maintenance_window_end,
                maintenance_spill,
            ) = get_start_date_end_date_from_time(
                base_date, maintenance_window[4:10] + maintenance_window[14::]
            )
            if (
                get_overlap_between_two_date_ranges(
                    backup_window_start,
                    backup_window_end,
                    maintenance_window_start,
                    maintenance_window_end,
                )
                >= 0
            ):
                return "The backup window and maintenance window must not overlap."

            # Due to spill overs, adjust the windows
            elif maintenance_spill:
                backup_window_start += datetime.timedelta(days=1)
                backup_window_end += datetime.timedelta(days=1)
            elif backup_spill:
                maintenance_window_start += datetime.timedelta(days=1)
                maintenance_window_end += datetime.timedelta(days=1)

            # If spills, rerun overlap test with adjusted windows
            if maintenance_spill or backup_spill:
                if (
                    get_overlap_between_two_date_ranges(
                        backup_window_start,
                        backup_window_end,
                        maintenance_window_start,
                        maintenance_window_end,
                    )
                    >= 0
                ):
                    return "The backup window and maintenance window must not overlap."

        maintenance_window_start, maintenance_window_end = get_start_date_end_date(
            base_date, maintenance_window
        )
        delta = maintenance_window_end - maintenance_window_start
        delta_seconds = delta.seconds + (delta.days * SECONDS_IN_ONE_DAY)
        if delta_seconds >= MINUTES_30 and delta_seconds <= HOURS_24:
            return None
        elif delta_seconds >= 0 and delta_seconds <= MINUTES_30:
            return "The maintenance window must be at least 30 minutes."
        else:
            return "Maintenance window must be less than 24 hours."
    except Exception:
        return f"Invalid day:hour:minute value: {maintenance_window}"


ORDERABLE_DB_INSTANCE_ENCODING = {
    "Engine": "E",
    "EngineVersion": "EV",
    "DBInstanceClass": "DBIC",
    "LicenseModel": "L",
    "AvailabilityZones": "AZ",
    "MultiAZCapable": "MC",
    "ReadReplicaCapable": "RC",
    "Vpc": "V",
    "SupportsStorageEncryption": "SE",
    "StorageType": "ST",
    "SupportsIops": "SI",
    "SupportsEnhancedMonitoring": "SM",
    "SupportsIAMDatabaseAuthentication": "SIAM",
    "SupportsPerformanceInsights": "SPI",
    "AvailableProcessorFeatures": "APF",
    "SupportedEngineModes": "SEM",
    "SupportsKerberosAuthentication": "SK",
    "OutpostCapable": "O",
    "SupportedActivityStreamModes": "SSM",
    "SupportsGlobalDatabases": "SGD",
    "SupportsClusters": "SC",
    "SupportedNetworkTypes": "SN",
    "SupportsStorageThroughput": "SST",
}
ORDERABLE_DB_INSTANCE_DECODING = {
    v: k for (k, v) in ORDERABLE_DB_INSTANCE_ENCODING.items()
}


def encode_orderable_db_instance(db: Dict[str, Any]) -> Dict[str, Any]:
    encoded = copy.deepcopy(db)
    if "AvailabilityZones" in encoded:
        encoded["AvailabilityZones"] = [
            az["Name"] for az in encoded["AvailabilityZones"]
        ]
    return {
        ORDERABLE_DB_INSTANCE_ENCODING.get(key, key): value
        for key, value in encoded.items()
    }


def decode_orderable_db_instance(db: Dict[str, Any]) -> Dict[str, Any]:
    decoded = copy.deepcopy(db)
    decoded_az = ORDERABLE_DB_INSTANCE_ENCODING.get(
        "AvailabilityZones", "AvailabilityZones"
    )
    if decoded_az in decoded:
        decoded["AvailabilityZones"] = [{"Name": az} for az in decoded[decoded_az]]
    return {
        ORDERABLE_DB_INSTANCE_DECODING.get(key, key): value
        for key, value in decoded.items()
    }


@lru_cache()
def get_service_model(service_name: str) -> ServiceModel:
    loader = create_loader()
    model = loader.load_service_model(service_name, "service-2")
    service_model = ServiceModel(model, service_name)
    return service_model
